"use strict";
/**
 * The commands for the quint CLI
 *
 * See the description at:
 * https://github.com/informalsystems/quint/blob/main/doc/quint.md
 *
 * @author Igor Konnov, Gabriela Moreira, Shon Feder, Informal Systems, 2021-2023
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.outputResult = exports.docs = exports.verifySpec = exports.outputCompilationTarget = exports.compile = exports.runSimulator = exports.runTests = exports.runRepl = exports.typecheck = exports.mkErrorMessage = exports.parse = exports.load = void 0;
const fs_1 = require("fs");
const json_bigint_1 = __importDefault(require("json-bigint"));
const path_1 = require("path");
const process_1 = require("process");
const chalk_1 = __importDefault(require("chalk"));
const quintParserFrontend_1 = require("./parsing/quintParserFrontend");
const assert_1 = require("assert");
const either_1 = require("@sweet-monads/either");
const repl_1 = require("./repl");
const quintIr_1 = require("./ir/quintIr");
const errorReporter_1 = require("./errorReporter");
const docs_1 = require("./docs");
const quintError_1 = require("./quintError");
const testing_1 = require("./runtime/testing");
const idGenerator_1 = require("./idGenerator");
const simulation_1 = require("./simulation");
const itf_1 = require("./itf");
const graphics_1 = require("./graphics");
const verbosity_1 = require("./verbosity");
const rng_1 = require("./rng");
const sourceResolver_1 = require("./parsing/sourceResolver");
const quintVerifier_1 = require("./quintVerifier");
const fullFlattener_1 = require("./flattening/fullFlattener");
const quintAnalyzer_1 = require("./quintAnalyzer");
const lodash_1 = require("lodash");
const maybe_1 = require("@sweet-monads/maybe");
const compileToTlaplus_1 = require("./compileToTlaplus");
// Extract just the parts of a ProcedureStage that we use for the output
// See https://stackoverflow.com/a/39333479/1187277
const pickOutputStage = ({ stage, warnings, modules, table, unusedDefinitions, types, effects, errors, documentation, passed, failed, ignored, status, trace, }) => {
    return {
        stage,
        warnings,
        modules,
        table,
        unusedDefinitions,
        types,
        effects,
        errors,
        documentation,
        passed,
        failed,
        ignored,
        status,
        trace,
    };
};
function cliErr(msg, stage) {
    return (0, either_1.left)({ msg, stage });
}
/** Load a file into a string
 *
 * @param args the CLI arguments parsed by yargs */
async function load(args) {
    const stage = { stage: 'loading', args };
    if ((0, fs_1.existsSync)(args.input)) {
        try {
            const path = (0, path_1.resolve)((0, process_1.cwd)(), args.input);
            const sourceCode = (0, fs_1.readFileSync)(path, 'utf8');
            return (0, either_1.right)({
                ...stage,
                args,
                path,
                sourceCode: new Map([[path, sourceCode]]),
                warnings: [],
            });
        }
        catch (err) {
            return cliErr(`file ${args.input} could not be opened due to ${err}`, {
                ...stage,
                errors: [],
                sourceCode: new Map(),
            });
        }
    }
    else {
        return cliErr(`file ${args.input} does not exist`, { ...stage, errors: [], sourceCode: new Map() });
    }
}
exports.load = load;
/**
 * Parse a Quint specification.
 *
 * @param loaded the procedure stage produced by `load`
 */
async function parse(loaded) {
    const { args, sourceCode, path } = loaded;
    const text = sourceCode.get(path);
    const parsing = { ...loaded, stage: 'parsing' };
    const idGen = (0, idGenerator_1.newIdGenerator)();
    return (0, lodash_1.flow)([
        () => {
            const phase1Data = (0, quintParserFrontend_1.parsePhase1fromText)(idGen, text, path);
            // if there is exactly one module in the original text, make it the main one
            const defaultModuleName = phase1Data.modules.length === 1 ? (0, maybe_1.just)(phase1Data.modules[0].name) : (0, maybe_1.none)();
            return { ...phase1Data, defaultModuleName };
        },
        phase1Data => {
            const resolver = (0, sourceResolver_1.fileSourceResolver)(sourceCode);
            const mainPath = resolver.lookupPath((0, path_1.dirname)(path), (0, path_1.basename)(path));
            return (0, quintParserFrontend_1.parsePhase2sourceResolution)(idGen, resolver, mainPath, phase1Data);
        },
        phase2Data => {
            if (args.sourceMap) {
                // Write source map to the specified file
                writeToJson(args.sourceMap, (0, quintParserFrontend_1.compactSourceMap)(phase2Data.sourceMap));
            }
            return (0, quintParserFrontend_1.parsePhase3importAndNameResolution)(phase2Data);
        },
        phase3Data => (0, quintParserFrontend_1.parsePhase4toposort)(phase3Data),
        phase4Data => ({ ...parsing, ...phase4Data, idGen }),
        result => {
            if (result.errors.length > 0) {
                const newErrorMessages = result.errors.map(mkErrorMessage(result.sourceMap));
                const errorMessages = parsing.errors ? parsing.errors.concat(newErrorMessages) : newErrorMessages;
                return (0, either_1.left)({ msg: 'parsing failed', stage: { ...result, errors: errorMessages } });
            }
            return (0, either_1.right)(result);
        },
    ])();
}
exports.parse = parse;
function mkErrorMessage(sourceMap) {
    return error => {
        const loc = error.reference ? sourceMap.get(error.reference) : undefined;
        return {
            explanation: (0, quintError_1.quintErrorToString)(error),
            locs: loc ? [loc] : [],
        };
    };
}
exports.mkErrorMessage = mkErrorMessage;
/**
 * Check types and effects of a Quint specification.
 *
 * @param parsed the procedure stage produced by `parse`
 */
async function typecheck(parsed) {
    const { table, modules, sourceMap } = parsed;
    const [errorMap, result] = (0, quintAnalyzer_1.analyzeModules)(table, modules);
    const typechecking = { ...parsed, ...result, stage: 'typechecking' };
    if (errorMap.length === 0) {
        return (0, either_1.right)(typechecking);
    }
    else {
        const errors = errorMap.map(mkErrorMessage(sourceMap));
        return cliErr('typechecking failed', { ...typechecking, errors });
    }
}
exports.typecheck = typecheck;
/**
 * Run REPL.
 *
 * @param _argv parameters as provided by yargs
 */
async function runRepl(_argv) {
    let filename = undefined;
    let moduleName = undefined;
    if (_argv.require) {
        // quint -r FILE.qnt or quint -r FILE.qnt::MODULE
        const m = /^(.*?)(?:|::([a-zA-Z_]\w*))$/.exec(_argv.require);
        if (m) {
            ;
            [filename, moduleName] = m.slice(1, 3);
        }
    }
    const options = {
        preloadFilename: filename,
        importModule: moduleName,
        verbosity: _argv.quiet ? 0 : _argv.verbosity,
    };
    (0, repl_1.quintRepl)(process.stdin, process.stdout, options);
}
exports.runRepl = runRepl;
/**
 * Run the tests. We imitate the output of mocha.
 *
 * @param typedStage the procedure stage produced by `typecheck`
 */
async function runTests(prev) {
    const testing = { ...prev, stage: 'testing' };
    const verbosityLevel = deriveVerbosity(prev.args);
    const mainName = guessMainModule(prev);
    const rngOrError = mkRng(prev.args.seed);
    if (rngOrError.isLeft()) {
        return cliErr(rngOrError.value, { ...testing, errors: [] });
    }
    const rng = rngOrError.unwrap();
    const matchFun = (n) => isMatchingTest(testing.args.match, n);
    const options = {
        testMatch: matchFun,
        maxSamples: testing.args.maxSamples,
        rng,
        verbosity: verbosityLevel,
        onTrace: (index, name, status, vars, states) => {
            if (outputTemplate && outputTemplate.endsWith('.itf.json')) {
                const filename = outputTemplate.replaceAll('{}', name).replaceAll('{#}', index);
                const trace = (0, itf_1.toItf)(vars, states);
                if (trace.isRight()) {
                    const jsonObj = addItfHeader(prev.args.input, status, trace.value);
                    writeToJson(filename, jsonObj);
                }
                else {
                    console.error(`ITF conversion failed on ${name}: ${trace.value}`);
                }
            }
        },
    };
    const out = console.log;
    const outputTemplate = testing.args.output;
    // Start the Timer and being running the tests
    const startMs = Date.now();
    if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
        out(`\n  ${mainName}`);
    }
    // TODO: This block is a workaround for the fact that flattening removes any defs
    // not refernced in the main module. We'd instead like way to just instruct it
    // to keep some defs.
    //
    // Find tests that are not used in the main module. We need to add references to them in the main module so flattening
    // doesn't drop the definitions.
    const unusedTests = [...testing.unusedDefinitions(mainName)].filter(d => d.kind === 'def' && options.testMatch(d.name));
    // Define name expressions referencing each test that is not referenced elsewhere, adding the references to the lookup
    // table
    const args = unusedTests.map(defToAdd => {
        const id = testing.idGen.nextId();
        testing.table.set(id, defToAdd);
        const namespace = defToAdd.importedFrom ? (0, quintIr_1.qualifier)(defToAdd.importedFrom) : undefined;
        const name = namespace ? [namespace, defToAdd.name].join('::') : defToAdd.name;
        return { kind: 'name', id, name };
    });
    // Wrap all expressions in a single declaration
    const testDeclaration = {
        id: testing.idGen.nextId(),
        name: 'q::unitTests',
        kind: 'def',
        qualifier: 'run',
        expr: { kind: 'app', opcode: 'actionAll', args, id: testing.idGen.nextId() },
    };
    // Add the declaration to the main module
    const main = testing.modules.find(m => m.name === mainName);
    main?.declarations.push(testDeclaration);
    // TODO The following block should all be moved into compileAndTest
    const analysisOutput = { types: testing.types, effects: testing.effects, modes: testing.modes };
    const { flattenedModules, flattenedTable, flattenedAnalysis } = (0, fullFlattener_1.flattenModules)(testing.modules, testing.table, testing.idGen, testing.sourceMap, analysisOutput);
    const compilationState = {
        originalModules: testing.modules,
        modules: flattenedModules,
        sourceMap: testing.sourceMap,
        analysisOutput: flattenedAnalysis,
        idGen: testing.idGen,
        sourceCode: testing.sourceCode,
    };
    const testResult = (0, testing_1.compileAndTest)(compilationState, mainName, flattenedTable, options);
    // We have a compilation failure, so early exit without reporting test results
    if (testResult.isLeft()) {
        return cliErr('Tests could not be run due to an error during compilation', {
            ...testing,
            errors: testResult.value.map(mkErrorMessage(testing.sourceMap)),
        });
    }
    // We're finished running the tests
    const elapsedMs = Date.now() - startMs;
    // So we can analyze the results now
    const results = testResult.value;
    // output the status for every test
    let nFailures = 1;
    if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
        results.forEach(res => {
            if (res.status === 'passed') {
                out(`    ${chalk_1.default.green('ok')} ${res.name} passed ${res.nsamples} test(s)`);
            }
            if (res.status === 'failed') {
                const errNo = chalk_1.default.red(nFailures);
                out(`    ${errNo}) ${res.name} failed after ${res.nsamples} test(s)`);
                nFailures++;
            }
        });
    }
    const passed = results.filter(r => r.status === 'passed');
    const failed = results.filter(r => r.status === 'failed');
    const ignored = results.filter(r => r.status === 'ignored');
    const namedErrors = failed.reduce((acc, failure) => acc.concat(failure.errors.map(e => [failure, mkErrorMessage(testing.sourceMap)(e)])), []);
    // output the statistics banner
    if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
        out('');
        if (passed.length > 0) {
            out(chalk_1.default.green(`  ${passed.length} passing`) + chalk_1.default.gray(` (${elapsedMs}ms)`));
        }
        if (failed.length > 0) {
            out(chalk_1.default.red(`  ${failed.length} failed`));
        }
        if (ignored.length > 0) {
            out(chalk_1.default.gray(`  ${ignored.length} ignored`));
        }
    }
    const stage = {
        ...testing,
        passed: passed.map(r => r.name),
        failed: failed.map(r => r.name),
        ignored: ignored.map(r => r.name),
        errors: namedErrors.map(([_, e]) => e),
    };
    // Nothing failed, so we are OK, and can exit early
    if (failed.length === 0) {
        return (0, either_1.right)(stage);
    }
    // We know that there are errors, so report as required by the verbosity configuration
    if (verbosity_1.verbosity.hasTestDetails(verbosityLevel)) {
        const code = prev.sourceCode;
        const finders = (0, errorReporter_1.createFinders)(code);
        const columns = !prev.args.out ? (0, graphics_1.terminalWidth)() : 80;
        out('');
        namedErrors.forEach(([testResult, err], index) => {
            const details = (0, errorReporter_1.formatError)(code, finders, err);
            // output the header
            out(`  ${index + 1}) ${testResult.name}:`);
            const lines = details.split('\n');
            // output the first two lines in red
            lines.slice(0, 2).forEach(l => out(chalk_1.default.red('      ' + l)));
            if (verbosity_1.verbosity.hasActionTracking(verbosityLevel)) {
                out('');
                testResult.frames.forEach((f, index) => {
                    out(`[${chalk_1.default.bold('Frame ' + index)}]`);
                    const console = {
                        width: columns,
                        out: (s) => process.stdout.write(s),
                    };
                    (0, graphics_1.printExecutionFrameRec)(console, f, []);
                    out('');
                });
                if (testResult.frames.length == 0) {
                    out('    [No execution]');
                }
            }
            // output the seed
            out(chalk_1.default.gray(`    Use --seed=0x${testResult.seed.toString(16)} --match=${testResult.name} to repeat.`));
        });
        out('');
    }
    if (verbosity_1.verbosity.hasHints(options.verbosity) && !verbosity_1.verbosity.hasActionTracking(options.verbosity)) {
        out(chalk_1.default.gray(`\n  Use --verbosity=3 to show executions.`));
        out(chalk_1.default.gray(`  Further debug with: quint --verbosity=3 ${prev.args.input}`));
    }
    return cliErr('Tests failed', stage);
}
exports.runTests = runTests;
// Print a counterexample if the appropriate verbosity is set
function maybePrintCounterExample(verbosityLevel, states, frames = []) {
    if (verbosity_1.verbosity.hasStateOutput(verbosityLevel)) {
        console.log(chalk_1.default.gray('An example execution:\n'));
        const myConsole = {
            width: (0, graphics_1.terminalWidth)(),
            out: (s) => process.stdout.write(s),
        };
        (0, graphics_1.printTrace)(myConsole, states, frames);
    }
}
/**
 * Run the simulator.
 *
 * @param prev the procedure stage produced by `typecheck`
 */
async function runSimulator(prev) {
    const simulator = { ...prev, stage: 'running' };
    const verbosityLevel = deriveVerbosity(prev.args);
    const mainName = guessMainModule(prev);
    const rngOrError = mkRng(prev.args.seed);
    if (rngOrError.isLeft()) {
        return cliErr(rngOrError.value, { ...simulator, errors: [] });
    }
    const rng = rngOrError.unwrap();
    const options = {
        init: prev.args.init,
        step: prev.args.step,
        invariant: prev.args.invariant,
        maxSamples: prev.args.maxSamples,
        maxSteps: prev.args.maxSteps,
        rng,
        verbosity: verbosityLevel,
    };
    const startMs = Date.now();
    const mainText = prev.sourceCode.get(prev.path);
    const mainPath = (0, sourceResolver_1.fileSourceResolver)(prev.sourceCode).lookupPath((0, path_1.dirname)(prev.args.input), (0, path_1.basename)(prev.args.input));
    const mainModule = prev.modules.find(m => m.name === mainName);
    if (mainModule === undefined) {
        return cliErr(`Main module ${mainName} not found`, { ...simulator, errors: [] });
    }
    const mainId = mainModule.id;
    const mainStart = prev.sourceMap.get(mainId).start.index;
    const mainEnd = prev.sourceMap.get(mainId).end.index;
    const result = (0, simulation_1.compileAndRun)((0, idGenerator_1.newIdGenerator)(), mainText, mainStart, mainEnd, mainName, mainPath, options);
    const elapsedMs = Date.now() - startMs;
    if (prev.args.outItf) {
        const trace = (0, itf_1.toItf)(result.vars, result.states);
        if (trace.isRight()) {
            const jsonObj = addItfHeader(prev.args.input, result.outcome.status, trace.value);
            writeToJson(prev.args.outItf, jsonObj);
        }
        else {
            return cliErr(`ITF conversion failed: ${trace.value}`, { ...simulator, errors: [] });
        }
    }
    switch (result.outcome.status) {
        case 'error':
            return cliErr('Runtime error', {
                ...simulator,
                status: result.outcome.status,
                trace: result.states,
                errors: result.outcome.errors.map(mkErrorMessage(prev.sourceMap)),
            });
        case 'ok':
            maybePrintCounterExample(verbosityLevel, result.states, result.frames);
            if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.green('[ok]') + ' No violation found ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                console.log(chalk_1.default.gray(`Use --seed=0x${result.seed.toString(16)} to reproduce.`));
                if (verbosity_1.verbosity.hasHints(options.verbosity)) {
                    console.log(chalk_1.default.gray('You may increase --max-samples and --max-steps.'));
                    console.log(chalk_1.default.gray('Use --verbosity to produce more (or less) output.'));
                }
            }
            return (0, either_1.right)({
                ...simulator,
                status: result.outcome.status,
                trace: result.states,
            });
        case 'violation':
            maybePrintCounterExample(verbosityLevel, result.states, result.frames);
            if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.red(`[violation]`) + ' Found an issue ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                console.log(chalk_1.default.gray(`Use --seed=0x${result.seed.toString(16)} to reproduce.`));
                if (verbosity_1.verbosity.hasHints(options.verbosity)) {
                    console.log(chalk_1.default.gray('Use --verbosity=3 to show executions.'));
                }
            }
            return cliErr('Invariant violated', {
                ...simulator,
                status: result.outcome.status,
                trace: result.states,
                errors: [],
            });
    }
}
exports.runSimulator = runSimulator;
/**  Compile to a flattened module, that includes the special q::* declarations
 *
 * @param typechecked the output of a preceding type checking stage
 */
async function compile(typechecked) {
    const args = typechecked.args;
    const mainName = guessMainModule(typechecked);
    const main = typechecked.modules.find(m => m.name === mainName);
    if (!main) {
        return cliErr(`module ${mainName} does not exist`, { ...typechecked, errors: [], sourceCode: new Map() });
    }
    // Wrap init, step, invariant and temporal properties in other definitions,
    // to make sure they are not considered unused in the main module and,
    // therefore, ignored by the flattener
    const extraDefsAsText = [`action q::init = ${args.init}`, `action q::step = ${args.step}`];
    if (args.invariant) {
        extraDefsAsText.push(`val q::inv = and(${args.invariant})`);
    }
    if (args.temporal) {
        extraDefsAsText.push(`temporal q::temporalProps = and(${args.temporal})`);
    }
    const extraDefs = extraDefsAsText.map(d => (0, quintParserFrontend_1.parseDefOrThrow)(d, typechecked.idGen, new Map()));
    main.declarations.push(...extraDefs);
    // We have to update the lookup table and analysis result with the new definitions. This is not ideal, and the problem
    // is that is hard to add this definitions in the proper stage, in our current setup. We should try to tackle this
    // while solving #1052.
    const resolutionResult = (0, quintParserFrontend_1.parsePhase3importAndNameResolution)({ ...typechecked, errors: [] });
    if (resolutionResult.errors.length > 0) {
        const errors = resolutionResult.errors.map(mkErrorMessage(typechecked.sourceMap));
        return cliErr('name resolution failed', { ...typechecked, errors });
    }
    typechecked.table = resolutionResult.table;
    (0, quintAnalyzer_1.analyzeInc)(typechecked, typechecked.table, extraDefs);
    // Flatten modules, replacing instances, imports and exports with their definitions
    const { flattenedModules, flattenedTable, flattenedAnalysis } = (0, fullFlattener_1.flattenModules)(typechecked.modules, typechecked.table, typechecked.idGen, typechecked.sourceMap, typechecked);
    // Pick the main module
    const flatMain = flattenedModules.find(m => m.name === mainName);
    return (0, either_1.right)({
        ...typechecked,
        ...flattenedAnalysis,
        mainModule: flatMain,
        table: flattenedTable,
        stage: 'compiling',
    });
}
exports.compile = compile;
/** output a compiled spec in the format specified in the `compiled.args.target` to stdout
 *
 * @param compiled The result of a preceding compile stage
 */
async function outputCompilationTarget(compiled) {
    const stage = 'outputting target';
    const args = compiled.args;
    const verbosityLevel = deriveVerbosity(args);
    const parsedSpecJson = jsonStringOfOutputStage(pickOutputStage({ ...compiled, modules: [compiled.mainModule] }));
    switch (compiled.args.target.toLowerCase()) {
        case 'json':
            process.stdout.write(parsedSpecJson);
            return (0, either_1.right)(compiled);
        case 'tlaplus': {
            const toTlaResult = await (0, compileToTlaplus_1.compileToTlaplus)(parsedSpecJson, verbosityLevel);
            return toTlaResult
                .mapRight(tla => {
                process.stdout.write(tla); // Write out, since all went right
                return compiled;
            })
                .mapLeft(err => {
                return {
                    msg: err.explanation,
                    stage: { ...compiled, stage, status: 'error', errors: err.errors },
                };
            });
        }
        default:
            // This is validated in the arg parsing
            (0, assert_1.fail)(`Invalid option for --target`);
    }
}
exports.outputCompilationTarget = outputCompilationTarget;
/**
 * Verify a spec via Apalache.
 *
 * @param prev the procedure stage produced by `typecheck`
 */
async function verifySpec(prev) {
    const verifying = { ...prev, stage: 'verifying' };
    const args = verifying.args;
    const verbosityLevel = deriveVerbosity(args);
    let loadedConfig = {};
    try {
        if (args.apalacheConfig) {
            loadedConfig = JSON.parse((0, fs_1.readFileSync)(args.apalacheConfig, 'utf-8'));
        }
    }
    catch (err) {
        return cliErr(`failed to read Apalache config: ${err.message}`, { ...verifying, errors: [], sourceCode: new Map() });
    }
    const veryfiyingFlat = { ...prev, modules: [prev.mainModule] };
    const parsedSpec = jsonStringOfOutputStage(pickOutputStage(veryfiyingFlat));
    // We need to insert the data form CLI args into their appropriate locations
    // in the Apalache config
    const config = {
        ...loadedConfig,
        input: {
            ...(loadedConfig.input ?? {}),
            source: {
                type: 'string',
                format: 'qnt',
                content: parsedSpec,
            },
        },
        checker: {
            ...(loadedConfig.checker ?? {}),
            length: args.maxSteps,
            init: 'q::init',
            next: 'q::step',
            inv: args.invariant ? ['q::inv'] : undefined,
            'temporal-props': args.temporal ? ['q::temporalProps'] : undefined,
            tuning: {
                ...(loadedConfig.checker?.tuning ?? {}),
                'search.simulation': args.randomTransitions ? 'true' : 'false',
            },
        },
    };
    const startMs = Date.now();
    return (0, quintVerifier_1.verify)(config, verbosityLevel).then(res => {
        const elapsedMs = Date.now() - startMs;
        return res
            .map(_ => {
            if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                console.log(chalk_1.default.green('[ok]') + ' No violation found ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                if (verbosity_1.verbosity.hasHints(verbosityLevel)) {
                    console.log(chalk_1.default.gray('You may increase --max-steps.'));
                    console.log(chalk_1.default.gray('Use --verbosity to produce more (or less) output.'));
                }
            }
            return { ...verifying, status: 'ok', errors: [] };
        })
            .mapLeft(err => {
            const trace = err.traces ? (0, itf_1.ofItf)(err.traces[0]) : undefined;
            const status = trace !== undefined ? 'violation' : 'failure';
            if (trace !== undefined) {
                // Always print the conterexample, unless the output is being directed to one of the outfiles
                maybePrintCounterExample(verbosityLevel, trace);
                if (verbosity_1.verbosity.hasResults(verbosityLevel)) {
                    console.log(chalk_1.default.red(`[${status}]`) + ' Found an issue ' + chalk_1.default.gray(`(${elapsedMs}ms).`));
                }
                if (prev.args.outItf) {
                    writeToJson(prev.args.outItf, err.traces);
                }
            }
            return {
                msg: err.explanation,
                stage: { ...verifying, status, errors: err.errors, trace },
            };
        });
    });
}
exports.verifySpec = verifySpec;
/**
 * Produces documentation from docstrings in a Quint specification.
 *
 * @param loaded the procedure stage produced by `load`
 */
async function docs(loaded) {
    const { sourceCode, path } = loaded;
    const text = sourceCode.get(path);
    const parsing = { ...loaded, stage: 'documentation' };
    const phase1Data = (0, quintParserFrontend_1.parsePhase1fromText)((0, idGenerator_1.newIdGenerator)(), text, path);
    const allEntries = phase1Data.modules.map(module => {
        const documentationEntries = (0, docs_1.produceDocs)(module);
        const title = `# Documentation for ${module.name}\n\n`;
        const markdown = title + [...documentationEntries.values()].map(docs_1.toMarkdown).join('\n\n');
        console.log(markdown);
        return [module.name, documentationEntries];
    });
    if (phase1Data.errors.length > 0) {
        const newErrorMessages = phase1Data.errors.map(mkErrorMessage(phase1Data.sourceMap));
        const errorMessages = parsing.errors ? parsing.errors.concat(newErrorMessages) : newErrorMessages;
        return (0, either_1.left)({ msg: 'parsing failed', stage: { ...parsing, errors: errorMessages } });
    }
    return (0, either_1.right)({ ...parsing, documentation: new Map(allEntries) });
}
exports.docs = docs;
/** Write the OutputStage of the procedureStage as JSON, if --out is set
 * Otherwise, report any stage errors to STDOUT
 */
function outputResult(result) {
    result
        .map(stage => {
        const outputData = pickOutputStage(stage);
        if (stage.args.out) {
            writeToJson(stage.args.out, outputData);
        }
        process.exit(0);
    })
        .mapLeft(({ msg, stage }) => {
        const { args, errors, sourceCode } = stage;
        const outputData = pickOutputStage(stage);
        if (args.out) {
            writeToJson(args.out, outputData);
        }
        else {
            const finders = (0, errorReporter_1.createFinders)(sourceCode);
            (0, lodash_1.uniqWith)(errors, lodash_1.isEqual).forEach(err => console.error((0, errorReporter_1.formatError)(sourceCode, finders, err)));
            console.error(`error: ${msg}`);
        }
        process.exit(1);
    });
}
exports.outputResult = outputResult;
function guessMainModule(stage) {
    if (stage.args.main) {
        // the main module is specified via --main
        return stage.args.main;
    }
    if (stage.defaultModuleName.isJust()) {
        // there is only one module in the source file, make it main
        return stage.defaultModuleName.unwrap();
    }
    // guess the name from the filename
    return (0, path_1.basename)(stage.args.input, '.qnt');
}
/**
 * Produce a random-number generator: Either a predictable one using a seed,
 * or a reasonably unpredictable one.
 */
function mkRng(seedText) {
    let seed;
    if (seedText !== undefined) {
        // since yargs does not has a type for big integers,
        // we do it with a fallback
        try {
            seed = BigInt(seedText);
        }
        catch (SyntaxError) {
            return (0, either_1.left)(`--seed must be a big integer, found: ${seedText}`);
        }
    }
    return (0, either_1.right)(seed ? (0, rng_1.newRng)(seed) : (0, rng_1.newRng)());
}
function addItfHeader(source, status, traceInJson) {
    return {
        '#meta': {
            format: 'ITF',
            'format-description': 'https://apalache.informal.systems/docs/adr/015adr-trace.html',
            source,
            status,
            description: 'Created by Quint on ' + new Date(),
            timestamp: Date.now(),
        },
        ...traceInJson,
    };
}
// Preprocess troublesome types so they are represented in JSON.
//
// We need it particularly because, by default, serialization of Map and Set
// objects just produces an empty object
// (see https://stackoverflow.com/questions/46634449/json-stringify-of-object-of-map-return-empty)
//
// The approach here follows https://stackoverflow.com/a/56150320/1187277
function replacer(_key, value) {
    if (value instanceof Map) {
        // Represent Maps as JSON objects
        return Object.fromEntries(value);
    }
    else if (value instanceof Set) {
        // Represent Sets as JSON arrays
        return Array.from(value);
    }
    else {
        return value;
    }
}
function jsonStringOfOutputStage(json) {
    return json_bigint_1.default.stringify(json, replacer);
}
/**
 * Write json to a file.
 *
 * @param filename name of the file to write to
 * @param json is an object tree to write
 */
function writeToJson(filename, json) {
    const path = (0, path_1.resolve)((0, process_1.cwd)(), filename);
    (0, fs_1.writeFileSync)(path, jsonStringOfOutputStage(json));
}
/**
 * Does a definition name match the expected test criteria.
 *
 * @param tests an optional array of test names
 * @param name the name of a definition to match
 * @returns whether the name matches the tests, if tests are not undefined,
 *          or name ends with 'Test'
 *
 */
function isMatchingTest(match, name) {
    if (match) {
        return new RegExp(match).exec(name) !== null;
    }
    else {
        return name.endsWith('Test');
    }
}
// Derive the verbosity for simulation and verification routines
function deriveVerbosity(args) {
    return !args.out && !args.outItf ? args.verbosity : 0;
}
//# sourceMappingURL=cliCommands.js.map