"use strict";
/* ----------------------------------------------------------------------------------
 * Copyright 2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 * --------------------------------------------------------------------------------- */
Object.defineProperty(exports, "__esModule", { value: true });
exports.stringSourceResolver = exports.fileSourceResolver = void 0;
/**
 * Source resolution for Quint. This module deals with loading sources from strings
 * and files.
 *
 * @author Igor Konnov
 *
 * @module
 */
const either_1 = require("@sweet-monads/either");
const path_1 = require("path");
const fs_1 = require("fs");
const path_2 = require("path");
const eol_1 = require("eol");
/**
 * Read the source code in UTF-8 from the filesystem via NodeJS API.
 * @param sourceCode an optional map of paths to source code,
 *        to be updated when reading new files
 * @param replacer an optional path replacement function,
 *        which is used to produce a source name
 * @returns A filesystem resolver. For each path, it returns
 *          either `left(errorMessage)`, or `right(fileContents)`.
 */
function fileSourceResolver(sourceCode = new Map(), replacer = path => path) {
    return {
        lookupPath: (basepath, importPath) => {
            return {
                normalizedPath: path_2.posix.join(basepath, importPath),
                toSourceName: () => {
                    return replacer(path_2.posix.join(basepath, importPath));
                },
            };
        },
        load: (lookupPath) => {
            try {
                const code = (0, eol_1.lf)((0, fs_1.readFileSync)(lookupPath.normalizedPath, 'utf8'));
                sourceCode.set(lookupPath.normalizedPath, code);
                return (0, either_1.right)(code);
            }
            catch (err) {
                return (0, either_1.left)(err.message);
            }
        },
        stempath: (lookupPath) => {
            return (0, path_1.dirname)(lookupPath.normalizedPath);
        },
    };
}
exports.fileSourceResolver = fileSourceResolver;
/**
 * Read the source code from a map of strings. This resolver is especially
 * useful for tests.
 * @param sources a map of paths mapped to text
 * @returns a static resolver that uses the map to read the contents.
 */
const stringSourceResolver = (sources) => {
    return {
        lookupPath: (stempath, importPath) => {
            return {
                normalizedPath: path_2.posix.join(stempath, importPath),
                toSourceName: () => {
                    return path_2.posix.join(stempath, importPath);
                },
            };
        },
        load: (lookupPath) => {
            // We are using nodejs path.join here.
            // If we have to decouple this resolver from nodejs in the future,
            // we would have to write our own version of join.
            const contents = sources.get(lookupPath.normalizedPath);
            return contents ? (0, either_1.right)(contents) : (0, either_1.left)(`Source not found: '${lookupPath.normalizedPath}'`);
        },
        stempath: (lookupPath) => {
            return (0, path_1.dirname)(lookupPath.normalizedPath);
        },
    };
};
exports.stringSourceResolver = stringSourceResolver;
//# sourceMappingURL=sourceResolver.js.map