"use strict";
/*
 * REPL for quint.
 *
 * Igor Konnov, Gabriela Moreira, 2022-2023.
 *
 * Copyright 2022-2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.quintRepl = exports.settings = void 0;
const readline = __importStar(require("readline"));
const fs_1 = require("fs");
const maybe_1 = require("@sweet-monads/maybe");
const either_1 = require("@sweet-monads/either");
const chalk_1 = __importDefault(require("chalk"));
const prettierimp_1 = require("./prettierimp");
const compile_1 = require("./runtime/compile");
const errorReporter_1 = require("./errorReporter");
const trace_1 = require("./runtime/trace");
const quintParserFrontend_1 = require("./parsing/quintParserFrontend");
const graphics_1 = require("./graphics");
const verbosity_1 = require("./verbosity");
const rng_1 = require("./rng");
const version_1 = require("./version");
const sourceResolver_1 = require("./parsing/sourceResolver");
const process_1 = require("process");
const idGenerator_1 = require("./idGenerator");
const IRprinting_1 = require("./ir/IRprinting");
const compilerImpl_1 = require("./runtime/impl/compilerImpl");
const cliCommands_1 = require("./cliCommands");
// tunable settings
exports.settings = {
    prompt: '>>> ',
    continuePrompt: '... ',
};
/**
 * The internal state of the REPL.
 */
class ReplState {
    constructor(verbosityLevel, rng) {
        this.moduleHist = '';
        this.exprHist = [];
        this.lastLoadedFileAndModule = [undefined, undefined];
        this.compilationState = (0, compile_1.newCompilationState)();
        this.evaluationState = (0, compilerImpl_1.newEvaluationState)((0, trace_1.newTraceRecorder)(verbosityLevel, rng));
    }
    clone() {
        const copy = new ReplState(this.verbosity, (0, rng_1.newRng)(this.rng.getState()));
        copy.moduleHist = this.moduleHist;
        copy.exprHist = this.exprHist;
        copy.lastLoadedFileAndModule = this.lastLoadedFileAndModule;
        copy.compilationState = this.compilationState;
        return copy;
    }
    addReplModule() {
        const replModule = { name: '__repl__', declarations: simulatorBuiltins(this.compilationState), id: 0n };
        this.compilationState.modules.push(replModule);
        this.compilationState.originalModules.push(replModule);
        this.compilationState.mainName = '__repl__';
        this.moduleHist += (0, IRprinting_1.moduleToString)(replModule);
    }
    clear() {
        this.moduleHist = '';
        this.exprHist = [];
        this.compilationState = (0, compile_1.newCompilationState)();
        this.evaluationState = (0, compilerImpl_1.newEvaluationState)((0, trace_1.newTraceRecorder)(this.verbosity, this.rng));
    }
    get recorder() {
        // ReplState always passes TraceRecorder in the evaluation state
        return this.evaluationState.listener;
    }
    get rng() {
        return this.recorder.rng;
    }
    get verbosity() {
        return this.recorder.verbosityLevel;
    }
    set verbosity(level) {
        this.recorder.verbosityLevel = level;
    }
    get seed() {
        return this.rng.getState();
    }
    set seed(newSeed) {
        this.rng.setState(newSeed);
    }
}
// The default exit terminates the process.
// Since it is inconvenient for testing, do not use it in tests :)
function defaultExit() {
    process.exit(0);
}
// the entry point to the REPL
function quintRepl(input, output, options = {
    verbosity: verbosity_1.verbosity.defaultLevel,
}, exit = defaultExit) {
    // output a line of text, no line feed is introduced
    const out = (text) => output.write(text);
    const prompt = (text) => {
        return verbosity_1.verbosity.hasReplPrompt(options.verbosity) ? text : '';
    };
    if (verbosity_1.verbosity.hasReplBanners(options.verbosity)) {
        out(chalk_1.default.gray(`Quint REPL ${version_1.version}\n`));
        out(chalk_1.default.gray('Type ".exit" to exit, or ".help" for more information\n'));
    }
    // create a readline interface
    const rl = readline.createInterface({
        input,
        output,
        prompt: prompt(exports.settings.prompt),
    });
    // the state
    const state = new ReplState(options.verbosity, (0, rng_1.newRng)());
    // we let the user type a multiline string, which is collected here:
    let multilineText = '';
    // when recyclingOwnOutput is true, REPL is receiving its older output
    let recyclingOwnOutput = false;
    // when the number of open braces or parentheses is positive,
    // we enter the multiline mode
    let nOpenBraces = 0;
    let nOpenParen = 0;
    let nOpenComments = 0;
    // Ctrl-C handler
    rl.on('SIGINT', () => {
        // if the user is stuck and presses Ctrl-C, reset the multiline mode
        multilineText = '';
        nOpenBraces = 0;
        nOpenParen = 0;
        nOpenComments = 0;
        rl.setPrompt(prompt(exports.settings.prompt));
        out(chalk_1.default.yellow(' <cancelled>\n'));
        // clear the line and show the prompt
        rl.write(null, { ctrl: true, name: 'u' });
        rl.prompt();
    });
    // next line handler
    function nextLine(line) {
        const [nob, nop, noc] = countBraces(line);
        nOpenBraces += nob;
        nOpenParen += nop;
        nOpenComments += noc;
        if (multilineText === '') {
            // if the line starts with a non-empty prompt,
            // we assume it is multiline code that was copied from a REPL prompt
            recyclingOwnOutput = exports.settings.prompt !== '' && line.trim().indexOf(exports.settings.prompt) === 0;
            if (nOpenBraces > 0 || nOpenParen > 0 || nOpenComments > 0 || recyclingOwnOutput) {
                // Enter a multiline mode.
                // If the text is copy-pasted from the REPL output,
                // trim the REPL decorations.
                multilineText = trimReplDecorations(line);
                rl.setPrompt(prompt(exports.settings.continuePrompt));
            }
            else {
                if (line.trim() !== '') {
                    tryEvalAndClearRecorder(out, state, line + '\n');
                }
            }
        }
        else {
            const trimmedLine = line.trim();
            const continueOwnOutput = exports.settings.continuePrompt !== '' && trimmedLine.indexOf(exports.settings.continuePrompt) === 0;
            if ((trimmedLine.length === 0 && nOpenBraces <= 0 && nOpenParen <= 0 && nOpenComments <= 0) ||
                (recyclingOwnOutput && !continueOwnOutput)) {
                // End the multiline mode.
                // If recycle own output, then the current line is, most likely,
                // older input. Ignore it.
                tryEvalAndClearRecorder(out, state, multilineText);
                multilineText = '';
                recyclingOwnOutput = false;
                rl.setPrompt(prompt(exports.settings.prompt));
            }
            else {
                // Continue the multiline mode.
                // It may happen that the text is copy-pasted from the REPL output.
                // In this case, we have to trim the leading '... '.
                multilineText += '\n' + trimReplDecorations(trimmedLine);
            }
        }
    }
    // load the code from a filename and optionally import a module
    function load(filename, moduleName) {
        state.clear();
        const newState = loadFromFile(out, state, filename);
        if (!newState) {
            return state;
        }
        state.lastLoadedFileAndModule[0] = filename;
        const moduleNameToLoad = moduleName ?? getMainModuleAnnotation(newState.moduleHist) ?? '__repl__';
        if (moduleNameToLoad === '__repl__') {
            // No module to load, introduce the __repl__ module
            newState.addReplModule();
        }
        if (tryEvalModule(out, newState, moduleNameToLoad)) {
            state.lastLoadedFileAndModule[1] = moduleNameToLoad;
        }
        else {
            out(chalk_1.default.yellow('Pick the right module name and import it (the file has been loaded)\n'));
            return newState;
        }
        if (newState.exprHist) {
            newState.exprHist.forEach(expr => {
                tryEvalAndClearRecorder(out, newState, expr);
            });
        }
        state.moduleHist = newState.moduleHist;
        state.exprHist = newState.exprHist;
        state.compilationState = newState.compilationState;
        state.evaluationState = newState.evaluationState;
    }
    // the read-eval-print loop
    rl.on('line', line => {
        const r = (s) => {
            return chalk_1.default.red(s);
        };
        const g = (s) => {
            return chalk_1.default.gray(s);
        };
        if (!line.startsWith('.')) {
            // an input to evaluate
            nextLine(line);
        }
        else {
            // a special command to REPL, extract the command name
            const m = line.match(/^\s*\.(\w+)/);
            if (m === null) {
                out(r(`Unexpected command: ${line}\n`));
                out(g(`Type .help to see the list of commands\n`));
            }
            else {
                switch (m[1]) {
                    case 'help':
                        out(`${r('.clear')}\tClear the history\n`);
                        out(`${r('.exit')}\tExit the REPL\n`);
                        out(`${r('.help')}\tPrint this help message\n`);
                        out(`${r('.load')} <filename> ${g('[<module>]')}\tClear the history,\n`);
                        out('     \tload the code from a file into the REPL session\n');
                        out('     \tand optionally import all definitions from <module>\n');
                        out(`${r('.reload')}\tClear the history, load and (optionally) import the last loaded file.\n`);
                        out('     \t^ a productivity hack\n');
                        out(`${r('.save')} <filename>\tSave the accumulated definitions to a file\n`);
                        out(`${r('.verbosity')}=[0-5]\tSet the output level (0 = quiet, 5 = very detailed).\n`);
                        out(`${r('.seed')}[=<number>]\tSet or get the random seed.\n`);
                        out('\nType an expression and press Enter to evaluate it.\n');
                        out('When the REPL switches to multiline mode "...", finish it with an empty line.\n');
                        out('\nPress Ctrl+C to abort current expression, Ctrl+D to exit the REPL\n');
                        break;
                    case 'exit':
                        exit();
                        break;
                    case 'clear':
                        out('\n'); // be nice to external programs
                        state.clear();
                        break;
                    case 'load':
                        {
                            const args = line.trim().split(/\s+/);
                            const [filename, moduleName] = [args[1], args[2]];
                            if (!filename) {
                                out(r('.load requires a filename\n'));
                            }
                            else {
                                load(filename, moduleName);
                            }
                            rl.prompt();
                        }
                        break;
                    case 'reload':
                        if (state.lastLoadedFileAndModule[0] !== undefined) {
                            load(state.lastLoadedFileAndModule[0], state.lastLoadedFileAndModule[1]);
                        }
                        else {
                            out(r('Nothing to reload. Use: .load filename [moduleName].\n'));
                        }
                        break;
                    case 'save':
                        {
                            const args = line.trim().split(/\s+/);
                            if (!args[1]) {
                                out(r('.save requires a filename\n'));
                            }
                            else {
                                saveToFile(out, state, args[1]);
                            }
                            rl.prompt();
                        }
                        break;
                    case 'verbosity':
                        {
                            // similar to yargs, accept: .verbosity n, .verbosity=n, .verbosity = n
                            const m = line.match(/^\.verbosity\s*=?\s*([0-5])$/);
                            if (m === null) {
                                out(r('.verbosity requires a level from 0 to 5\n'));
                            }
                            else {
                                state.verbosity = Number(m[1]);
                                if (verbosity_1.verbosity.hasReplPrompt(state.verbosity)) {
                                    out(g(`.verbosity=${state.verbosity}\n`));
                                }
                                rl.setPrompt(prompt(exports.settings.prompt));
                            }
                        }
                        break;
                    case 'seed':
                        {
                            // accept: .seed n, .seed=n, .seed = n
                            const m = line.match(/^\.seed\s*=?\s*((0x[0-9a-f]+|[0-9]*))$/);
                            if (m === null) {
                                out(r('.seed requires an integer, or no argument\n'));
                            }
                            else {
                                if (m[1].trim() === '') {
                                    out(g(`.seed=${state.seed}\n`));
                                }
                                else {
                                    state.seed = BigInt(m[1]);
                                    if (verbosity_1.verbosity.hasReplPrompt(state.verbosity)) {
                                        out(g(`.seed=${state.seed}\n`));
                                    }
                                }
                            }
                        }
                        break;
                    default:
                        out(r(`Unexpected command: ${line}\n`));
                        out(g(`Type .help to see the list of commands\n`));
                        break;
                }
            }
        }
        rl.prompt();
    }).on('close', () => {
        out('\n');
        exit();
    });
    // Everything is registered. Optionally, load a module.
    if (options.preloadFilename) {
        load(options.preloadFilename, options.importModule);
    }
    rl.prompt();
    return rl;
}
exports.quintRepl = quintRepl;
function saveToFile(out, state, filename) {
    // 1. Write the previously loaded modules.
    // 2. Write the definitions in the special module called __repl__.
    // 3. Wrap expressions into special comments.
    try {
        const text = `// @mainModule ${state.lastLoadedFileAndModule[1]}\n` +
            `${state.moduleHist}` +
            state.exprHist.map(s => `/*! ${s} !*/\n`).join('\n');
        (0, fs_1.writeFileSync)(filename, text);
        out(`Session saved to: ${filename}\n`);
    }
    catch (error) {
        out(chalk_1.default.red(error));
        out('\n');
    }
}
function loadFromFile(out, state, filename) {
    try {
        const modules = (0, fs_1.readFileSync)(filename, 'utf8');
        const newState = state.clone();
        newState.moduleHist = modules + newState.moduleHist;
        // unwrap the expressions from the specially crafted comments
        const exprs = Array.from((modules ?? '').matchAll(/\/\*! (.*?) !\*\//gms) ?? []).map(groups => groups[1]);
        // and replay them one by one
        newState.exprHist = exprs;
        return newState;
    }
    catch (error) {
        out(chalk_1.default.red(error));
        out('\n');
        return;
    }
}
/**
 * Moves the nextvars register values into the vars, and clears the nextvars.
 * Returns an array of variable names that were not updated.
 * @param vars An array of Register objects representing the current state of the variables.
 * @param nextvars An array of Register objects representing the next state of the variables.
 * @returns An array of variable names that were not updated, or none if all variables were updated.
 */
function saveVars(vars, nextvars) {
    let isAction = false;
    const nonUpdated = vars.reduce((acc, varRegister) => {
        const nextVarRegister = nextvars.find(v => v.name === varRegister.name);
        if (nextVarRegister && nextVarRegister.registerValue.isJust()) {
            varRegister.registerValue = nextVarRegister.registerValue;
            nextVarRegister.registerValue = (0, maybe_1.none)();
            isAction = true;
        }
        else {
            // No nextvar for this variable, so it was not updated
            acc.push(varRegister.name);
        }
        return acc;
    }, []);
    if (isAction) {
        // return the names of the variables that have not been updated
        return (0, maybe_1.just)(nonUpdated);
    }
    else {
        return (0, maybe_1.none)();
    }
}
// Declarations that are overloaded by the simulator.
// In the future, we will declare them in a separate module.
function simulatorBuiltins(st) {
    return [
        (0, quintParserFrontend_1.parseDefOrThrow)(`def q::test = (q::nruns, q::nsteps, q::init, q::next, q::inv) => false`, st.idGen, st.sourceMap),
        (0, quintParserFrontend_1.parseDefOrThrow)(`def q::testOnce = (q::nsteps, q::init, q::next, q::inv) => false`, st.idGen, st.sourceMap),
    ];
}
function tryEvalModule(out, state, mainName) {
    const modulesText = state.moduleHist;
    const mainPath = (0, sourceResolver_1.fileSourceResolver)(state.compilationState.sourceCode).lookupPath((0, process_1.cwd)(), 'repl.ts');
    state.compilationState.sourceCode.set(mainPath.toSourceName(), modulesText);
    const context = (0, compile_1.compileFromCode)((0, idGenerator_1.newIdGenerator)(), modulesText, mainName, mainPath, state.evaluationState.listener, state.rng.next);
    if (context.evaluationState?.context.size === 0 ||
        context.compileErrors.length > 0 ||
        context.syntaxErrors.length > 0) {
        const tempState = state.clone();
        // The compilation state has updated source code maps, to be used in error reporting
        tempState.compilationState = context.compilationState;
        printErrors(out, tempState, context);
        return false;
    }
    if (context.analysisErrors.length > 0) {
        printErrors(out, state, context);
        // provisionally, continue on type & effects errors
    }
    // Save compilation state
    state.compilationState = context.compilationState;
    state.evaluationState = context.evaluationState;
    return true;
}
// Try to evaluate the expression in a string and print it, if successful.
// After that, clear the recorded stack.
function tryEvalAndClearRecorder(out, state, newInput) {
    const result = tryEval(out, state, newInput);
    state.recorder.clear();
    return result;
}
// try to evaluate the expression in a string and print it, if successful
function tryEval(out, state, newInput) {
    if (state.compilationState.modules.length === 0) {
        state.addReplModule();
        tryEvalModule(out, state, '__repl__');
    }
    const parseResult = (0, quintParserFrontend_1.parseExpressionOrDeclaration)(newInput, '<input>', state.compilationState.idGen, state.compilationState.sourceMap);
    if (parseResult.kind === 'error') {
        printErrorMessages(out, state, 'syntax error', newInput, parseResult.errors);
        out('\n'); // be nice to external programs
        return false;
    }
    if (parseResult.kind === 'none') {
        // a comment or whitespaces
        return true;
    }
    // evaluate the input, depending on its type
    if (parseResult.kind === 'expr') {
        const context = (0, compile_1.compileExpr)(state.compilationState, state.evaluationState, state.rng, parseResult.expr);
        if (context.syntaxErrors.length > 0 || context.compileErrors.length > 0 || context.analysisErrors.length > 0) {
            printErrors(out, state, context, newInput);
            if (context.syntaxErrors.length > 0 || context.compileErrors.length > 0) {
                return false;
            } // else: provisionally, continue on type & effects errors
        }
        state.exprHist.push(newInput.trim());
        // Save the evaluation state only, as state vars changes should persist
        state.evaluationState = context.evaluationState;
        return evalExpr(state, out)
            .mapLeft(msg => {
            // when #618 is implemented, we should remove this
            printErrorMessages(out, state, 'runtime error', newInput, context.getRuntimeErrors());
            // print the error message produced by the lookup
            out(chalk_1.default.red(msg));
            out('\n'); // be nice to external programs
        })
            .isRight();
    }
    if (parseResult.kind === 'declaration') {
        // compile the module and add it to history if everything worked
        const context = (0, compile_1.compileDecls)(state.compilationState, state.evaluationState, state.rng, parseResult.decls);
        if (context.evaluationState.context.size === 0 ||
            context.compileErrors.length > 0 ||
            context.syntaxErrors.length > 0) {
            printErrors(out, state, context, newInput);
            return false;
        }
        if (context.analysisErrors.length > 0) {
            printErrors(out, state, context, newInput);
            // provisionally, continue on type & effects errors
        }
        out('\n'); // be nice to external programs
        state.moduleHist = state.moduleHist.slice(0, state.moduleHist.length - 1) + newInput + '\n}'; // update the history
        // Save compilation state
        state.compilationState = context.compilationState;
        state.evaluationState = context.evaluationState;
    }
    return true;
}
// output errors to the console in red
function printErrors(out, state, context, newInput = '') {
    printErrorMessages(out, state, 'syntax error', newInput, context.syntaxErrors);
    printErrorMessages(out, state, 'static analysis error', newInput, context.analysisErrors, chalk_1.default.yellow);
    printErrorMessages(out, state, 'compile error', newInput, context.compileErrors);
    out('\n'); // be nice to external programs
}
// print error messages with proper colors
function printErrorMessages(out, state, kind, inputText, errors, color = chalk_1.default.red) {
    const modulesText = state.moduleHist + inputText;
    const messages = errors.map((0, cliCommands_1.mkErrorMessage)(state.compilationState.sourceMap));
    // display the error messages and highlight the error places
    // FIXME(#1052): moudulesText can come from multiple files, but `compileFromCode` ignores that.
    // We use a fallback here to '<modules>'
    const sourceCode = new Map([
        ['<input>', inputText],
        ['<modules>', modulesText],
        ...state.compilationState.sourceCode.entries(),
    ]);
    const finders = (0, errorReporter_1.createFinders)(sourceCode);
    messages.forEach(e => {
        const error = {
            ...e,
            locs: e.locs.map(loc => ({ ...loc, source: finders.has(loc.source) ? loc.source : '<modules>' })),
        };
        const msg = (0, errorReporter_1.formatError)(sourceCode, finders, error, (0, maybe_1.none)());
        out(color(`${kind}: ${msg}\n`));
    });
}
// if a line start with '>>> ' or '... ', trim these markers
function trimReplDecorations(line) {
    // we are not using settings.prompt and settings.continuePrompt,
    // as ... are interpreted as three characters.
    const match = /^\s*(>>> |\.\.\. )(.*)/.exec(line);
    if (match && match[2] !== undefined) {
        return match[2];
    }
    else {
        return line;
    }
}
// count the difference between the number of:
//  - '{' and '}'
//  - '(' and ')'
//  - '/*' and '*/'
function countBraces(str) {
    let nOpenBraces = 0;
    let nOpenParen = 0;
    let nOpenComments = 0;
    for (let i = 0; i < str.length; i++) {
        switch (str[i]) {
            case '{':
                nOpenBraces++;
                break;
            case '}':
                nOpenBraces--;
                break;
            case '(':
                nOpenParen++;
                break;
            case ')':
                nOpenParen--;
                break;
            case '/':
                if (i + 1 < str.length && str[i + 1] === '*') {
                    nOpenComments++;
                    i++;
                }
                break;
            case '*':
                if (i + 1 < str.length && str[i + 1] === '/') {
                    nOpenComments--;
                    i++;
                }
                break;
            default:
        }
    }
    return [nOpenBraces, nOpenParen, nOpenComments];
}
function evalExpr(state, out) {
    const computable = (0, compile_1.contextNameLookup)(state.evaluationState.context, compile_1.inputDefName, 'callable');
    const columns = (0, graphics_1.terminalWidth)();
    const result = computable
        .mapRight(comp => {
        return comp
            .eval()
            .map(value => {
            const ex = value.toQuintEx(state.compilationState.idGen);
            out((0, prettierimp_1.format)(columns, 0, (0, graphics_1.prettyQuintEx)(ex)));
            out('\n');
            if (ex.kind === 'bool' && ex.value) {
                // A Boolean expression may be an action or a run.
                // Save the state, if there were any updates to variables.
                saveVars(state.evaluationState.vars, state.evaluationState.nextVars).map(missing => {
                    if (missing.length > 0) {
                        out(chalk_1.default.yellow('[warning] some variables are undefined: ' + missing.join(', ') + '\n'));
                    }
                });
            }
            return (0, either_1.right)(ex);
        })
            .or((0, maybe_1.just)((0, either_1.left)('<undefined value>')))
            .unwrap();
    })
        .join();
    if (verbosity_1.verbosity.hasUserOpTracking(state.verbosity)) {
        const trace = state.recorder.getBestTrace();
        if (trace.subframes.length > 0) {
            out('\n');
            trace.subframes.forEach((f, i) => {
                out(`[Frame ${i}]\n`);
                (0, graphics_1.printExecutionFrameRec)({ width: columns, out }, f, []);
                out('\n');
            });
        }
    }
    return result;
}
function getMainModuleAnnotation(moduleHist) {
    const moduleName = moduleHist.match(/^\/\/ @mainModule\s+(\w+)\n/);
    return moduleName?.at(1);
}
//# sourceMappingURL=repl.js.map