import { Either } from '@sweet-monads/either';
import { SourceMap } from '../parsing/quintParserFrontend';
import { Computable, ComputableKind } from './runtime';
import { ExecutionListener } from './trace';
import { FlatModule, QuintDeclaration, QuintDef, QuintEx, QuintModule } from '../ir/quintIr';
import { EvaluationState } from './impl/compilerImpl';
import { LookupTable } from '../names/base';
import { AnalysisOutput } from '../quintAnalyzer';
import { IdGenerator } from '../idGenerator';
import { SourceLookupPath } from '../parsing/sourceResolver';
import { Rng } from '../rng';
import { QuintError } from '../quintError';
/**
 * The name of the builtin name that returns the last found trace.
 */
export declare const lastTraceName = "q::lastTrace";
/**
 * The name of a definition that wraps the user input, e.g., in REPL.
 */
export declare const inputDefName: string;
/**
 * A compilation context returned by 'compile'.
 */
export interface CompilationContext {
    lookupTable: LookupTable;
    syntaxErrors: QuintError[];
    analysisErrors: QuintError[];
    compileErrors: QuintError[];
    getRuntimeErrors: () => QuintError[];
    compilationState: CompilationState;
    evaluationState: EvaluationState;
}
/**
 * A data structure that holds the state of the compilation process.
 */
export interface CompilationState {
    idGen: IdGenerator;
    sourceCode: Map<string, string>;
    originalModules: QuintModule[];
    modules: FlatModule[];
    mainName?: string;
    sourceMap: SourceMap;
    analysisOutput: AnalysisOutput;
}
export declare function newCompilationState(): CompilationState;
export declare function errorContextFromMessage(listener: ExecutionListener): (_: {
    errors: QuintError[];
    sourceMap: SourceMap;
}) => CompilationContext;
export declare function contextNameLookup(context: Map<string, Computable>, defName: string, kind: ComputableKind): Either<string, Computable>;
/**
 * Compile Quint defs to JS runtime objects from the parsed and type-checked
 * data structures. This is a user-facing function. In case of an error, the
 * error messages are passed to an error handler and the function returns
 * undefined.
 *
 * @param compilationState the state of the compilation process
 * @param evaluationState the state of the compiler visitor
 * @param lookupTable lookup table as produced by the parser
 * @param rand the random number generator
 * @param defs the definitions to compile
 * @returns the compilation context
 */
export declare function compile(compilationState: CompilationState, evaluationState: EvaluationState, lookupTable: LookupTable, rand: (bound: bigint) => bigint, defs: QuintDef[]): CompilationContext;
/**
 * Compile a single Quint expression, given a non-empty compilation and
 * evaluation state. That is, those states should have the results of the
 * compilation of at least one module.
 *
 * @param state - The current compilation state
 * @param evaluationState - The current evaluation state
 * @param rng - The random number generator
 * @param expr - The Quint exporession to be compiled
 *
 * @returns A compilation context with the compiled expression or its errors
 */
export declare function compileExpr(state: CompilationState, evaluationState: EvaluationState, rng: Rng, expr: QuintEx): CompilationContext;
/**
 * Compile a single Quint definition, given a non-empty compilation and
 * evaluation state. That is, those states should have the results of the
 * compilation of at least one module.
 *
 * @param state - The current compilation state
 * @param evaluationState - The current evaluation state
 * @param rng - The random number generator
 * @param decls - The Quint declarations to be compiled
 *
 * @returns A compilation context with the compiled definition or its errors
 */
export declare function compileDecls(state: CompilationState, evaluationState: EvaluationState, rng: Rng, decls: QuintDeclaration[]): CompilationContext;
/**
 * Parse a string that contains Quint modules and compile it to executable
 * objects. This is a user-facing function. In case of an error, the error
 * messages are passed to an error handler and the function returns undefined.
 *
 * @param code text that stores one or several Quint modules,
 *        which should be parseable without any context
 * @param mainName the name of the module that may contain state varibles
 * @param execListener execution listener
 * @param rand the random number generator
 * @returns the compilation context
 */
export declare function compileFromCode(idGen: IdGenerator, code: string, mainName: string, mainPath: SourceLookupPath, execListener: ExecutionListener, rand: (bound: bigint) => bigint): CompilationContext;
