"use strict";
/*
 * A compiler to the runtime environment.
 *
 * Igor Konnov, Informal Systems, 2022-2023
 *
 * Copyright 2022-2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileFromCode = exports.compileDecls = exports.compileExpr = exports.compile = exports.contextNameLookup = exports.errorContextFromMessage = exports.newCompilationState = exports.inputDefName = exports.lastTraceName = void 0;
const either_1 = require("@sweet-monads/either");
const quintParserFrontend_1 = require("../parsing/quintParserFrontend");
const runtime_1 = require("./runtime");
const compilerImpl_1 = require("./impl/compilerImpl");
const IRVisitor_1 = require("../ir/IRVisitor");
const quintAnalyzer_1 = require("../quintAnalyzer");
const idGenerator_1 = require("../idGenerator");
const fullFlattener_1 = require("../flattening/fullFlattener");
/**
 * The name of the builtin name that returns the last found trace.
 */
exports.lastTraceName = 'q::lastTrace';
/**
 * The name of a definition that wraps the user input, e.g., in REPL.
 */
exports.inputDefName = 'q::input';
/* An empty initial compilation state */
function newCompilationState() {
    return {
        idGen: (0, idGenerator_1.newIdGenerator)(),
        sourceCode: new Map(),
        originalModules: [],
        modules: [],
        sourceMap: new Map(),
        analysisOutput: {
            types: new Map(),
            effects: new Map(),
            modes: new Map(),
        },
    };
}
exports.newCompilationState = newCompilationState;
function errorContextFromMessage(listener) {
    const evaluationState = (0, compilerImpl_1.newEvaluationState)(listener);
    return ({ errors, sourceMap }) => {
        return {
            lookupTable: new Map(),
            syntaxErrors: errors,
            analysisErrors: [],
            compileErrors: [],
            getRuntimeErrors: () => [],
            compilationState: { ...newCompilationState(), sourceMap },
            evaluationState,
        };
    };
}
exports.errorContextFromMessage = errorContextFromMessage;
function contextNameLookup(context, defName, kind) {
    const value = context.get((0, runtime_1.kindName)(kind, defName));
    if (!value) {
        console.log(`key = ${(0, runtime_1.kindName)(kind, defName)}`);
        return (0, either_1.left)(`No value for definition ${defName}`);
    }
    else {
        return (0, either_1.right)(value);
    }
}
exports.contextNameLookup = contextNameLookup;
/**
 * Compile Quint defs to JS runtime objects from the parsed and type-checked
 * data structures. This is a user-facing function. In case of an error, the
 * error messages are passed to an error handler and the function returns
 * undefined.
 *
 * @param compilationState the state of the compilation process
 * @param evaluationState the state of the compiler visitor
 * @param lookupTable lookup table as produced by the parser
 * @param rand the random number generator
 * @param defs the definitions to compile
 * @returns the compilation context
 */
function compile(compilationState, evaluationState, lookupTable, rand, defs) {
    const visitor = new compilerImpl_1.CompilerVisitor(lookupTable, rand, evaluationState);
    defs.forEach(def => (0, IRVisitor_1.walkDefinition)(visitor, def));
    return {
        lookupTable,
        syntaxErrors: [],
        analysisErrors: [],
        compileErrors: visitor.getCompileErrors(),
        getRuntimeErrors: () => {
            return visitor.getRuntimeErrors().splice(0);
        },
        compilationState,
        evaluationState: visitor.getEvaluationState(),
    };
}
exports.compile = compile;
/**
 * Compile a single Quint expression, given a non-empty compilation and
 * evaluation state. That is, those states should have the results of the
 * compilation of at least one module.
 *
 * @param state - The current compilation state
 * @param evaluationState - The current evaluation state
 * @param rng - The random number generator
 * @param expr - The Quint exporession to be compiled
 *
 * @returns A compilation context with the compiled expression or its errors
 */
function compileExpr(state, evaluationState, rng, expr) {
    // Create a definition to encapsulate the parsed expression.
    // Note that the expression may contain nested definitions.
    // Hence, we have to compile it via an auxilliary definition.
    const def = { kind: 'def', qualifier: 'action', name: exports.inputDefName, expr, id: state.idGen.nextId() };
    return compileDecls(state, evaluationState, rng, [def]);
}
exports.compileExpr = compileExpr;
/**
 * Compile a single Quint definition, given a non-empty compilation and
 * evaluation state. That is, those states should have the results of the
 * compilation of at least one module.
 *
 * @param state - The current compilation state
 * @param evaluationState - The current evaluation state
 * @param rng - The random number generator
 * @param decls - The Quint declarations to be compiled
 *
 * @returns A compilation context with the compiled definition or its errors
 */
function compileDecls(state, evaluationState, rng, decls) {
    if (state.originalModules.length === 0 || state.modules.length === 0) {
        throw new Error('No modules in state');
    }
    // Define a new module list with the new definition in the main module,
    // ensuring the original object is not modified
    const originalModules = state.originalModules.map(m => {
        if (m.name === state.mainName) {
            return { ...m, declarations: [...m.declarations, ...decls] };
        }
        return m;
    });
    const mainModule = state.modules.find(m => m.name === state.mainName);
    // We need to resolve names for this new definition. Incremental name
    // resolution is not our focus now, so just resolve everything again.
    const { table, errors } = (0, quintParserFrontend_1.parsePhase3importAndNameResolution)({
        modules: originalModules,
        sourceMap: state.sourceMap,
        errors: [],
    });
    if (errors.length > 0) {
        // For now, don't try to run analysis and flattening if there are errors
        return errorContextFromMessage(evaluationState.listener)({ errors, sourceMap: state.sourceMap });
    }
    const [analysisErrors, analysisOutput] = (0, quintAnalyzer_1.analyzeInc)(state.analysisOutput, table, decls);
    const { flattenedModules: flatModules, flattenedTable, flattenedAnalysis, } = (0, fullFlattener_1.flattenModules)(originalModules, table, state.idGen, state.sourceMap, analysisOutput);
    const newState = {
        ...state,
        analysisOutput: flattenedAnalysis,
        modules: flatModules,
        originalModules: originalModules,
    };
    const flatDefinitions = flatModules.find(m => m.name === state.mainName).declarations;
    // Filter definitions that were not compiled yet
    const defsToCompile = flatDefinitions.filter(d => !mainModule.declarations.some(d2 => d2.id === d.id));
    const ctx = compile(newState, evaluationState, flattenedTable, rng.next, defsToCompile);
    return { ...ctx, analysisErrors };
}
exports.compileDecls = compileDecls;
/**
 * Parse a string that contains Quint modules and compile it to executable
 * objects. This is a user-facing function. In case of an error, the error
 * messages are passed to an error handler and the function returns undefined.
 *
 * @param code text that stores one or several Quint modules,
 *        which should be parseable without any context
 * @param mainName the name of the module that may contain state varibles
 * @param execListener execution listener
 * @param rand the random number generator
 * @returns the compilation context
 */
function compileFromCode(idGen, code, mainName, mainPath, execListener, rand) {
    // parse the module text
    // FIXME(#1052): We should build a proper sourceCode map from the files we previously loaded
    const sourceCode = new Map();
    const { modules, table, sourceMap, errors } = (0, quintParserFrontend_1.parse)(idGen, mainPath.toSourceName(), mainPath, code, sourceCode);
    // On errors, we'll produce the computational context up to this point
    const [analysisErrors, analysisOutput] = (0, quintAnalyzer_1.analyzeModules)(table, modules);
    const { flattenedModules, flattenedTable, flattenedAnalysis } = (0, fullFlattener_1.flattenModules)(modules, table, idGen, sourceMap, analysisOutput);
    const compilationState = {
        originalModules: modules,
        modules: flattenedModules,
        mainName,
        sourceMap,
        analysisOutput: flattenedAnalysis,
        idGen,
        sourceCode,
    };
    const main = flattenedModules.find(m => m.name === mainName);
    // when the main module is not found, we will report an error
    const mainNotFoundError = main
        ? []
        : [
            {
                code: 'QNT405',
                message: `Main module ${mainName} not found`,
            },
        ];
    const defsToCompile = main ? main.declarations : [];
    const ctx = compile(compilationState, (0, compilerImpl_1.newEvaluationState)(execListener), flattenedTable, rand, defsToCompile);
    return {
        ...ctx,
        compileErrors: ctx.compileErrors.concat(mainNotFoundError),
        analysisErrors,
        syntaxErrors: errors,
    };
}
exports.compileFromCode = compileFromCode;
//# sourceMappingURL=compile.js.map