import { LookupTable } from '../../names/base';
import { IRVisitor } from '../../ir/IRVisitor';
import { Computable, Register } from '../runtime';
import { ExecutionListener } from '../trace';
import * as ir from '../../ir/quintIr';
import { Trace } from './trace';
import { ErrorCode, QuintError } from '../../quintError';
/**
 * Returns a Map containing the built-in Computable objects for the Quint language.
 * These include the callable objects for Bool, Int, and Nat.
 *
 * @returns a Map containing the built-in Computable objects.
 */
export declare function builtinContext(): Map<string, Computable>;
/**
 * Represents the state of evaluation of Quint code.
 * All the fields are mutated by CompilerVisitor, either directly, or via calls.
 */
export interface EvaluationState {
    context: Map<string, Computable>;
    vars: Register[];
    nextVars: Register[];
    trace: Trace;
    errorTracker: CompilerErrorTracker;
    listener: ExecutionListener;
}
/**
 * Creates a new EvaluationState object.
 *
 * @returns a new EvaluationState object.
 */
export declare class CompilerErrorTracker {
    compileErrors: QuintError[];
    runtimeErrors: QuintError[];
    addCompileError(reference: bigint, code: ErrorCode, message: string): void;
    addRuntimeError(reference: bigint, code: ErrorCode, message: string): void;
}
/**
 * Creates a new EvaluationState object with the initial state of the evaluation.
 *
 * @returns a new EvaluationState object
 */
export declare function newEvaluationState(listener: ExecutionListener): EvaluationState;
/**
 * Compiler visitor turns Quint definitions and expressions into Computable
 * objects, essentially, lazy JavaScript objects. Importantly, it does not do
 * any evaluation during the translation and thus delegates the actual
 * computation to the JavaScript engine. Since many of Quint operators may be
 * computationally expensive, it is crucial to maintain this separation of
 * compilation vs. computation.
 *
 * This class does not do any dynamic type checking, assuming that the type
 * checker will be run before the translation in the future. As we do not have
 * the type checker yet, computations may fail with weird JavaScript errors.
 */
export declare class CompilerVisitor implements IRVisitor {
    private lookupTable;
    private compStack;
    context: Map<string, Computable>;
    private vars;
    private nextVars;
    private errorTracker;
    private rand;
    private execListener;
    private trace;
    definitionDepth: number;
    constructor(lookupTable: LookupTable, rand: (bound: bigint) => bigint, evaluationState: EvaluationState);
    /**
     * Get the compiler state.
     */
    getEvaluationState(): EvaluationState;
    /**
     * Get the names of the compiled variables.
     */
    getVars(): string[];
    /**
     * Get the array of compile errors, which changes as the code gets executed.
     */
    getCompileErrors(): QuintError[];
    /**
     * Get the array of runtime errors, which changes as the code gets executed.
     */
    getRuntimeErrors(): QuintError[];
    exitOpDef(opdef: ir.QuintOpDef): void;
    exitLet(letDef: ir.QuintLet): void;
    exitConst(cdef: ir.QuintConst): void;
    exitVar(vardef: ir.QuintVar): void;
    enterLiteral(expr: ir.QuintBool | ir.QuintInt | ir.QuintStr): void;
    enterName(name: ir.QuintName): void;
    exitApp(app: ir.QuintApp): void;
    private applyUserDefined;
    enterLambda(lam: ir.QuintLambda): void;
    exitLambda(lam: ir.QuintLambda): void;
    private translateAssign;
    /**
     * A generalized application of a one-argument Callable to a set-like
     * runtime value, as required by `exists`, `forall`, `map`, and `filter`.
     *
     * This method expects `compStack` to look like follows:
     *
     *  - `(top)` translated lambda, as `Callable`.
     *
     *  - `(top - 1)`: a set-like value to iterate over, as `Computable`.
     *
     * The method evaluates the Callable for each element of the iterable value
     * and either produces `none`, if evaluation failed for one of the elements,
     * or it applies `mapResultAndElems` to the pairs that consists of the Callable
     * result and the original element of the iterable value.
     * The final result is stored on the stack.
     */
    private mapLambdaThenReduce;
    /**
     * Translate one of the operators: fold, foldl, and foldr.
     */
    private applyFold;
    private applyFun;
    private translateIfThenElse;
    /**
     *  Compute all { ... } or A.then(B)...then(E) for a chain of actions.
     * @param actions actions as computable to execute
     * @param kind is it 'all { ... }' or 'A.then(B)'?
     * @param actionId given the action index, return the id that produced this action
     * @returns evaluation result
     */
    private chainAllOrThen;
    private translateAllOrThen;
    private translateExpect;
    private translateReps;
    private translateBoolOp;
    private translateOrAction;
    private applyOneOf;
    private test;
    private testOnce;
    private runTestSimulation;
    private varsToRecord;
    private shiftVars;
    private snapshotVars;
    private snapshotNextVars;
    private recoverVars;
    private recoverNextVars;
    private contextGet;
    private contextLookup;
    private getListElem;
    private updateList;
    private sliceList;
}
