"use strict";
/*
 * Compiler of Quint expressions and definitions to Computable values
 * that can be evaluated in the runtime.
 *
 * Igor Konnov, Gabriela Moreira, 2022-2023
 *
 * Copyright 2022-2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CompilerVisitor = exports.newEvaluationState = exports.CompilerErrorTracker = exports.builtinContext = void 0;
const assert_1 = require("assert");
const maybe_1 = require("@sweet-monads/maybe");
const immutable_1 = require("immutable");
const runtime_1 = require("../runtime");
const ir = __importStar(require("../../ir/quintIr"));
const runtimeValue_1 = require("./runtimeValue");
const trace_1 = require("./trace");
const compile_1 = require("../compile");
const graphics_1 = require("../../graphics");
const prettierimp_1 = require("../../prettierimp");
const util_1 = require("../../util");
const idGenerator_1 = require("../../idGenerator");
const lodash_1 = require("lodash");
// Internal names in the compiler, which have special treatment.
// For some reason, if we replace 'q::input' with inputDefName, everything breaks.
// What kind of JS magic is that?
const specialNames = ['q::input', 'q::runResult', 'q::nruns', 'q::nsteps', 'q::init', 'q::next', 'q::inv'];
/**
 * Returns a Map containing the built-in Computable objects for the Quint language.
 * These include the callable objects for Bool, Int, and Nat.
 *
 * @returns a Map containing the built-in Computable objects.
 */
function builtinContext() {
    return new Map([
        [(0, runtime_1.kindName)('callable', 'Bool'), (0, runtime_1.mkCallable)([], mkConstComputable(runtimeValue_1.rv.mkSet([runtimeValue_1.rv.mkBool(false), runtimeValue_1.rv.mkBool(true)])))],
        [(0, runtime_1.kindName)('callable', 'Int'), (0, runtime_1.mkCallable)([], mkConstComputable(runtimeValue_1.rv.mkInfSet('Int')))],
        [(0, runtime_1.kindName)('callable', 'Nat'), (0, runtime_1.mkCallable)([], mkConstComputable(runtimeValue_1.rv.mkInfSet('Nat')))],
    ]);
}
exports.builtinContext = builtinContext;
/**
 * Creates a new EvaluationState object.
 *
 * @returns a new EvaluationState object.
 */
class CompilerErrorTracker {
    constructor() {
        // messages that are produced during compilation
        this.compileErrors = [];
        // messages that get populated as the compiled code is executed
        this.runtimeErrors = [];
    }
    addCompileError(reference, code, message) {
        this.compileErrors.push({ code, message, reference });
    }
    addRuntimeError(reference, code, message) {
        this.runtimeErrors.push({ code, message, reference });
    }
}
exports.CompilerErrorTracker = CompilerErrorTracker;
/**
 * Creates a new EvaluationState object with the initial state of the evaluation.
 *
 * @returns a new EvaluationState object
 */
function newEvaluationState(listener) {
    const state = {
        context: builtinContext(),
        vars: [],
        nextVars: [],
        trace: new trace_1.Trace(),
        errorTracker: new CompilerErrorTracker(),
        listener: listener,
    };
    return state;
}
exports.newEvaluationState = newEvaluationState;
/**
 * Compiler visitor turns Quint definitions and expressions into Computable
 * objects, essentially, lazy JavaScript objects. Importantly, it does not do
 * any evaluation during the translation and thus delegates the actual
 * computation to the JavaScript engine. Since many of Quint operators may be
 * computationally expensive, it is crucial to maintain this separation of
 * compilation vs. computation.
 *
 * This class does not do any dynamic type checking, assuming that the type
 * checker will be run before the translation in the future. As we do not have
 * the type checker yet, computations may fail with weird JavaScript errors.
 */
class CompilerVisitor {
    constructor(lookupTable, rand, evaluationState) {
        // the stack of computable values
        this.compStack = [];
        // the current depth of operator definitions: top-level defs are depth 0
        // FIXME(#1279): The walk* functions update this value, but they need to be
        // initialized to -1 here for that to work on all scenarios.
        this.definitionDepth = -1;
        this.lookupTable = lookupTable;
        this.rand = rand;
        this.context = evaluationState.context;
        this.vars = evaluationState.vars;
        this.nextVars = evaluationState.nextVars;
        this.errorTracker = evaluationState.errorTracker;
        this.execListener = evaluationState.listener;
        this.trace = evaluationState.trace;
    }
    /**
     * Get the compiler state.
     */
    getEvaluationState() {
        return {
            context: this.context,
            vars: this.vars,
            nextVars: this.nextVars,
            errorTracker: this.errorTracker,
            listener: this.execListener,
            trace: this.trace,
        };
    }
    /**
     * Get the names of the compiled variables.
     */
    getVars() {
        return this.vars.map(r => r.name);
    }
    /**
     * Get the array of compile errors, which changes as the code gets executed.
     */
    getCompileErrors() {
        return this.errorTracker.compileErrors;
    }
    /**
     * Get the array of runtime errors, which changes as the code gets executed.
     */
    getRuntimeErrors() {
        return this.errorTracker.runtimeErrors;
    }
    exitOpDef(opdef) {
        // Either a runtime value, or a def, action, etc.
        // All of them are compiled to callables, which may have zero parameters.
        let boundValue = this.compStack.pop();
        if (boundValue === undefined) {
            this.errorTracker.addCompileError(opdef.id, 'QNT501', `No expression for ${opdef.name} on compStack`);
            return;
        }
        if (opdef.qualifier === 'action' && opdef.expr.kind !== 'lambda') {
            // A nullary action like `init` or `step`.
            // It is not handled via applyUserDefined.
            // Wrap this value with the listener calls.
            // Importantly, we do not touch the original boundValue, but decorate it.
            // Consider the following definitions:
            //   action input1 = step
            //   action input2 = step
            //
            // Both input1 and input2 wrap step, but in their individual computables.
            const unwrappedValue = boundValue;
            const app = { id: opdef.id, kind: 'app', opcode: opdef.name, args: [] };
            const evalApp = { id: 0n, kind: 'app', opcode: '_', args: [app] };
            boundValue = {
                eval: () => {
                    if (app.opcode === compile_1.inputDefName) {
                        this.execListener.onUserOperatorCall(evalApp);
                        // do not call onUserOperatorReturn on '_' later, as it may span over multiple frames
                    }
                    else {
                        this.execListener.onUserOperatorCall(app);
                    }
                    const r = unwrappedValue.eval();
                    this.execListener.onUserOperatorReturn(app, [], r);
                    return r;
                },
            };
        }
        if (this.definitionDepth === 0 && opdef.qualifier === 'pureval') {
            // a pure value may be cached, once evaluated
            const originalEval = boundValue.eval;
            let cache = undefined;
            boundValue.eval = () => {
                if (cache !== undefined) {
                    return cache;
                }
                else {
                    cache = originalEval();
                    return cache;
                }
            };
        }
        const kname = (0, runtime_1.kindName)('callable', opdef.id);
        // bind the callable from the stack
        this.context.set(kname, boundValue);
        if (specialNames.includes(opdef.name)) {
            // bind the callable under its name as well
            this.context.set((0, runtime_1.kindName)('callable', opdef.name), boundValue);
        }
    }
    exitLet(letDef) {
        // When dealing with a `val` or `nondet`, freeze the callable under
        // the let definition. Otherwise, forms like 'nondet x = oneOf(S); A'
        // may produce multiple values for the same name 'x'
        // inside a single evaluation of A.
        // In case of `val`, this is simply an optimization.
        const qualifier = letDef.opdef.qualifier;
        if (qualifier !== 'val' && qualifier !== 'nondet') {
            // a non-constant value, ignore
            return;
        }
        // get the expression that is evaluated in the context of let.
        const exprUnderLet = this.compStack.slice(-1).pop();
        if (exprUnderLet === undefined) {
            this.errorTracker.addCompileError(letDef.opdef.id, 'QNT501', `No expression for ${letDef.opdef.name} on compStack`);
            return;
        }
        const kname = (0, runtime_1.kindName)('callable', letDef.opdef.id);
        const boundValue = this.context.get(kname) ?? runtime_1.fail;
        // Override the behavior of the expression under let:
        // It precomputes the bound value and uses it in the evaluation.
        // Once the evaluation is done, the value is reset, so that
        // a new random value may be produced later.
        const undecoratedEval = exprUnderLet.eval;
        const boundValueEval = boundValue.eval;
        exprUnderLet.eval = function () {
            const cachedValue = boundValueEval();
            boundValue.eval = function () {
                return cachedValue;
            };
            // compute the result and immediately reset the cache
            const result = undecoratedEval();
            boundValue.eval = boundValueEval;
            return result;
        };
    }
    exitConst(cdef) {
        // all constants should be instantiated before running the simulator
        const code = 'QNT500';
        const msg = `Uninitialized const ${cdef.name}. Use: import <moduleName>(${cdef.name}=<value>).*`;
        this.errorTracker.addCompileError(cdef.id, code, msg);
    }
    exitVar(vardef) {
        const varName = vardef.name;
        // In the process of incremental compilation, we might revisit the same var
        // definition. Don't overwrite the register if that happens. In some cases
        // (with instances), the variable can have a different ID, but the same
        // name. In that case, we assign the register with that name for the new ID.
        if (this.context.has((0, runtime_1.kindName)('var', varName))) {
            const register = this.context.get((0, runtime_1.kindName)('var', varName));
            this.context.set((0, runtime_1.kindName)('var', vardef.id), register);
            if (this.context.has((0, runtime_1.kindName)('nextvar', varName))) {
                const register = this.context.get((0, runtime_1.kindName)('nextvar', varName));
                this.context.set((0, runtime_1.kindName)('nextvar', vardef.id), register);
            }
            return;
        }
        // simply introduce two registers:
        //  one for the variable, and
        //  one for its next-state version
        const prevRegister = (0, runtime_1.mkRegister)('var', varName, (0, maybe_1.none)(), () => this.errorTracker.addRuntimeError(vardef.id, 'QNT502', `Variable ${varName} is not set`));
        this.vars.push(prevRegister);
        // at the moment, we have to refer to variables both via id and name
        this.context.set((0, runtime_1.kindName)('var', varName), prevRegister);
        this.context.set((0, runtime_1.kindName)('var', vardef.id), prevRegister);
        const nextRegister = (0, runtime_1.mkRegister)('nextvar', varName, (0, maybe_1.none)(), () => this.errorTracker.addRuntimeError(vardef.id, 'QNT502', `${varName}' is not set`));
        this.nextVars.push(nextRegister);
        // at the moment, we have to refer to variables both via id and name
        this.context.set((0, runtime_1.kindName)('nextvar', varName), nextRegister);
        this.context.set((0, runtime_1.kindName)('nextvar', vardef.id), nextRegister);
    }
    enterLiteral(expr) {
        switch (expr.kind) {
            case 'bool':
                this.compStack.push(mkConstComputable(runtimeValue_1.rv.mkBool(expr.value)));
                break;
            case 'int':
                this.compStack.push(mkConstComputable(runtimeValue_1.rv.mkInt(expr.value)));
                break;
            case 'str':
                this.compStack.push(mkConstComputable(runtimeValue_1.rv.mkStr(expr.value)));
                break;
            default:
                (0, util_1.unreachable)(expr);
        }
    }
    enterName(name) {
        if (name.name === compile_1.lastTraceName) {
            this.compStack.push(mkConstComputable(runtimeValue_1.rv.mkList(this.trace.get())));
            return;
        }
        // The name belongs to one of the objects:
        // a shadow variable, a variable, an argument, a callable.
        // The order is important, as defines the name priority.
        const comp = this.contextLookup(name.id, ['arg', 'var', 'callable']) ??
            // a backup case for Nat, Int, and Bool, and special names such as q::input
            this.contextGet(name.name, ['arg', 'callable']);
        if (comp) {
            // this name has an associated computable object already
            this.compStack.push(comp);
        }
        else {
            // this should not happen, due to the name resolver
            this.errorTracker.addCompileError(name.id, 'QNT502', `Name ${name.name} not found`);
            this.compStack.push(runtime_1.fail);
        }
    }
    exitApp(app) {
        if (!ir.isQuintBuiltin(app)) {
            this.applyUserDefined(app);
        }
        else {
            switch (app.opcode) {
                case 'next':
                    {
                        const register = this.compStack.pop();
                        if (register) {
                            const name = register.name;
                            const nextvar = this.contextGet(name, ['nextvar']);
                            this.compStack.push(nextvar ?? runtime_1.fail);
                        }
                        else {
                            this.errorTracker.addCompileError(app.id, 'QNT502', 'Operator next(...) needs one argument');
                            this.compStack.push(runtime_1.fail);
                        }
                    }
                    break;
                case 'assign':
                    this.translateAssign(app.id);
                    break;
                case 'eq':
                    this.applyFun(app.id, 2, (x, y) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(x.equals(y))));
                    break;
                case 'neq':
                    this.applyFun(app.id, 2, (x, y) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(!x.equals(y))));
                    break;
                // conditional
                case 'ite':
                    this.translateIfThenElse(app.id);
                    break;
                // Booleans
                case 'not':
                    this.applyFun(app.id, 1, p => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(!p.toBool())));
                    break;
                case 'and':
                    // a conjunction over expressions is lazy
                    this.translateBoolOp(app, (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true)), (_, r) => (!r ? (0, maybe_1.just)(runtimeValue_1.rv.mkBool(false)) : (0, maybe_1.none)()));
                    break;
                case 'actionAll':
                    this.translateAllOrThen(app);
                    break;
                case 'or':
                    // a disjunction over expressions is lazy
                    this.translateBoolOp(app, (0, maybe_1.just)(runtimeValue_1.rv.mkBool(false)), (_, r) => (r ? (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true)) : (0, maybe_1.none)()));
                    break;
                case 'actionAny':
                    this.translateOrAction(app);
                    break;
                case 'implies':
                    // an implication is lazy
                    this.translateBoolOp(app, (0, maybe_1.just)(runtimeValue_1.rv.mkBool(false)), (n, r) => (n == 0 && !r ? (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true)) : (0, maybe_1.none)()));
                    break;
                case 'iff':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(p.toBool() === q.toBool())));
                    break;
                // integers
                case 'iuminus':
                    this.applyFun(app.id, 1, n => (0, maybe_1.just)(runtimeValue_1.rv.mkInt(-n.toInt())));
                    break;
                case 'iadd':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkInt(p.toInt() + q.toInt())));
                    break;
                case 'isub':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkInt(p.toInt() - q.toInt())));
                    break;
                case 'imul':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkInt(p.toInt() * q.toInt())));
                    break;
                case 'idiv':
                    this.applyFun(app.id, 2, (p, q) => {
                        if (q.toInt() !== 0n) {
                            return (0, maybe_1.just)(runtimeValue_1.rv.mkInt(p.toInt() / q.toInt()));
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT503', 'Division by zero');
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'imod':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkInt(p.toInt() % q.toInt())));
                    break;
                case 'ipow':
                    this.applyFun(app.id, 2, (p, q) => {
                        if (q.toInt() == 0n && p.toInt() == 0n) {
                            this.errorTracker.addRuntimeError(app.id, 'QNT503', '0^0 is undefined');
                        }
                        else if (q.toInt() < 0n) {
                            this.errorTracker.addRuntimeError(app.id, 'QNT503', 'i^j is undefined for j < 0');
                        }
                        else {
                            return (0, maybe_1.just)(runtimeValue_1.rv.mkInt(p.toInt() ** q.toInt()));
                        }
                        return (0, maybe_1.none)();
                    });
                    break;
                case 'igt':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(p.toInt() > q.toInt())));
                    break;
                case 'ilt':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(p.toInt() < q.toInt())));
                    break;
                case 'igte':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(p.toInt() >= q.toInt())));
                    break;
                case 'ilte':
                    this.applyFun(app.id, 2, (p, q) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(p.toInt() <= q.toInt())));
                    break;
                case 'Tup':
                    // Construct a tuple from an array of values
                    this.applyFun(app.id, app.args.length, (...values) => (0, maybe_1.just)(runtimeValue_1.rv.mkTuple(values)));
                    break;
                case 'item':
                    // Access a tuple: tuples are 1-indexed, that is, _1, _2, etc.
                    this.applyFun(app.id, 2, (tuple, idx) => this.getListElem(app.id, tuple.toList(), Number(idx.toInt()) - 1));
                    break;
                case 'tuples':
                    // Construct a cross product
                    this.applyFun(app.id, app.args.length, (...sets) => (0, maybe_1.just)(runtimeValue_1.rv.mkCrossProd(sets)));
                    break;
                case 'List':
                    // Construct a list from an array of values
                    this.applyFun(app.id, app.args.length, (...values) => (0, maybe_1.just)(runtimeValue_1.rv.mkList(values)));
                    break;
                case 'range':
                    this.applyFun(app.id, 2, (start, end) => {
                        const [s, e] = [Number(start.toInt()), Number(end.toInt())];
                        if (s <= e) {
                            const arr = [];
                            for (let i = s; i < e; i++) {
                                arr.push(runtimeValue_1.rv.mkInt(BigInt(i)));
                            }
                            return (0, maybe_1.just)(runtimeValue_1.rv.mkList(arr));
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT504', `range(${s}, ${e}) is out of bounds`);
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'nth':
                    // Access a list
                    this.applyFun(app.id, 2, (list, idx) => this.getListElem(app.id, list.toList(), Number(idx.toInt())));
                    break;
                case 'replaceAt':
                    this.applyFun(app.id, 3, (list, idx, value) => this.updateList(app.id, list.toList(), Number(idx.toInt()), value));
                    break;
                case 'head':
                    this.applyFun(app.id, 1, list => this.getListElem(app.id, list.toList(), 0));
                    break;
                case 'tail':
                    this.applyFun(app.id, 1, list => {
                        const l = list.toList();
                        if (l.size > 0) {
                            return this.sliceList(app.id, l, 1, l.size);
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT505', 'Applied tail to an empty list');
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'slice':
                    this.applyFun(app.id, 3, (list, start, end) => {
                        const [l, s, e] = [list.toList(), Number(start.toInt()), Number(end.toInt())];
                        if (s >= 0 && s <= l.size && e <= l.size && e >= s) {
                            return this.sliceList(app.id, l, s, e);
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT506', `slice(..., ${s}, ${e}) applied to a list of size ${l.size}`);
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'length':
                    this.applyFun(app.id, 1, list => (0, maybe_1.just)(runtimeValue_1.rv.mkInt(BigInt(list.toList().size))));
                    break;
                case 'append':
                    this.applyFun(app.id, 2, (list, elem) => (0, maybe_1.just)(runtimeValue_1.rv.mkList(list.toList().push(elem))));
                    break;
                case 'concat':
                    this.applyFun(app.id, 2, (list1, list2) => (0, maybe_1.just)(runtimeValue_1.rv.mkList(list1.toList().concat(list2.toList()))));
                    break;
                case 'indices':
                    this.applyFun(app.id, 1, list => (0, maybe_1.just)(runtimeValue_1.rv.mkInterval(0n, BigInt(list.toList().size - 1))));
                    break;
                case 'Rec':
                    // Construct a record
                    this.applyFun(app.id, app.args.length, (...values) => {
                        const keys = values.filter((e, i) => i % 2 === 0).map(k => k.toStr());
                        const map = keys.reduce((map, key, i) => {
                            const v = values[2 * i + 1];
                            return v ? map.set(key, v) : map;
                        }, (0, immutable_1.OrderedMap)());
                        return (0, maybe_1.just)(runtimeValue_1.rv.mkRecord(map));
                    });
                    break;
                case 'field':
                    // Access a record via the field name
                    this.applyFun(app.id, 2, (rec, fieldName) => {
                        const name = fieldName.toStr();
                        const fieldValue = rec.toOrderedMap().get(name);
                        if (fieldValue) {
                            return (0, maybe_1.just)(fieldValue);
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT501', `Accessing a missing record field ${name}`);
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'fieldNames':
                    this.applyFun(app.id, 1, rec => {
                        const keysAsRuntimeValues = rec
                            .toOrderedMap()
                            .keySeq()
                            .map(key => runtimeValue_1.rv.mkStr(key));
                        return (0, maybe_1.just)(runtimeValue_1.rv.mkSet(keysAsRuntimeValues));
                    });
                    break;
                case 'with':
                    // update a record
                    this.applyFun(app.id, 3, (rec, fieldName, fieldValue) => {
                        const oldMap = rec.toOrderedMap();
                        const key = fieldName.toStr();
                        if (oldMap.has(key)) {
                            const newMap = rec.toOrderedMap().set(key, fieldValue);
                            return (0, maybe_1.just)(runtimeValue_1.rv.mkRecord(newMap));
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT501', `Called 'with' with a non-existent key ${key}`);
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'variant':
                    // Construct a variant of a sum type.
                    this.applyFun(app.id, 2, (labelName, value) => (0, maybe_1.just)(runtimeValue_1.rv.mkVariant(labelName.toStr(), value)));
                    break;
                case 'matchVariant':
                    this.applyFun(app.id, app.args.length, (variantExpr, ...cases) => {
                        // Type checking ensures that this is a variant expression
                        (0, assert_1.strict)(variantExpr instanceof runtimeValue_1.RuntimeValueVariant, 'invalid value in match expression');
                        const label = variantExpr.label;
                        const value = variantExpr.value;
                        // Find the eliminator marked with the variant's label
                        let result;
                        for (const [caseLabel, caseElim] of (0, lodash_1.chunk)(cases, 2)) {
                            const caseLabelStr = caseLabel.toStr();
                            if (caseLabelStr === '_' || caseLabelStr === label) {
                                // Type checking ensures the second item of each case is a lambda
                                const eliminator = caseElim;
                                result = eliminator.eval([(0, maybe_1.just)(value)]).map(r => r);
                                break;
                            }
                        }
                        // Type checking ensures we have cases for every possible variant of a sum type.
                        (0, assert_1.strict)(result, 'non-exhaustive match expression');
                        return result;
                    });
                    break;
                case 'Set':
                    // Construct a set from an array of values.
                    this.applyFun(app.id, app.args.length, (...values) => (0, maybe_1.just)(runtimeValue_1.rv.mkSet(values)));
                    break;
                case 'powerset':
                    this.applyFun(app.id, 1, (baseset) => (0, maybe_1.just)(runtimeValue_1.rv.mkPowerset(baseset)));
                    break;
                case 'contains':
                    this.applyFun(app.id, 2, (set, value) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(set.contains(value))));
                    break;
                case 'in':
                    this.applyFun(app.id, 2, (value, set) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(set.contains(value))));
                    break;
                case 'subseteq':
                    this.applyFun(app.id, 2, (l, r) => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(l.isSubset(r))));
                    break;
                case 'union':
                    this.applyFun(app.id, 2, (l, r) => (0, maybe_1.just)(runtimeValue_1.rv.mkSet(l.toSet().union(r.toSet()))));
                    break;
                case 'intersect':
                    this.applyFun(app.id, 2, (l, r) => (0, maybe_1.just)(runtimeValue_1.rv.mkSet(l.toSet().intersect(r.toSet()))));
                    break;
                case 'exclude':
                    this.applyFun(app.id, 2, (l, r) => (0, maybe_1.just)(runtimeValue_1.rv.mkSet(l.toSet().subtract(r.toSet()))));
                    break;
                case 'size':
                    this.applyFun(app.id, 1, set => set.cardinality().map(runtimeValue_1.rv.mkInt));
                    break;
                case 'isFinite':
                    // at the moment, we support only finite sets, so just return true
                    this.applyFun(app.id, 1, _set => (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true)));
                    break;
                case 'to':
                    this.applyFun(app.id, 2, (i, j) => (0, maybe_1.just)(runtimeValue_1.rv.mkInterval(i.toInt(), j.toInt())));
                    break;
                case 'fold':
                    this.applyFold(app.id, 'fwd');
                    break;
                case 'foldl':
                    this.applyFold(app.id, 'fwd');
                    break;
                case 'foldr':
                    this.applyFold(app.id, 'rev');
                    break;
                case 'flatten':
                    this.applyFun(app.id, 1, set => {
                        // unpack the sets from runtime values
                        const setOfSets = set.toSet().map(e => e.toSet());
                        // and flatten the set of sets via immutable-js
                        return (0, maybe_1.just)(runtimeValue_1.rv.mkSet(setOfSets.flatten(1)));
                    });
                    break;
                case 'get':
                    // Get a map value
                    this.applyFun(app.id, 2, (map, key) => {
                        const value = map.toMap().get(key.normalForm());
                        if (value) {
                            return (0, maybe_1.just)(value);
                        }
                        else {
                            // Should we print the key? It may be a complex expression.
                            this.errorTracker.addRuntimeError(app.id, 'QNT507', "Called 'get' with a non-existing key");
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'set':
                    // Update a map value
                    this.applyFun(app.id, 3, (map, key, newValue) => {
                        const normalKey = key.normalForm();
                        const asMap = map.toMap();
                        if (asMap.has(normalKey)) {
                            const newMap = asMap.set(normalKey, newValue);
                            return (0, maybe_1.just)(runtimeValue_1.rv.fromMap(newMap));
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT507', "Called 'set' with a non-existing key");
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                case 'put':
                    // add a value to a map
                    this.applyFun(app.id, 3, (map, key, newValue) => {
                        const normalKey = key.normalForm();
                        const asMap = map.toMap();
                        const newMap = asMap.set(normalKey, newValue);
                        return (0, maybe_1.just)(runtimeValue_1.rv.fromMap(newMap));
                    });
                    break;
                case 'setBy': {
                    // Update a map value via a lambda
                    const fun = this.compStack.pop() ?? runtime_1.fail;
                    this.applyFun(app.id, 2, (map, key) => {
                        const normalKey = key.normalForm();
                        const asMap = map.toMap();
                        if (asMap.has(normalKey)) {
                            return fun.eval([(0, maybe_1.just)(asMap.get(normalKey))]).map(newValue => {
                                const newMap = asMap.set(normalKey, newValue);
                                return runtimeValue_1.rv.fromMap(newMap);
                            });
                        }
                        else {
                            this.errorTracker.addRuntimeError(app.id, 'QNT507', "Called 'setBy' with a non-existing key");
                            return (0, maybe_1.none)();
                        }
                    });
                    break;
                }
                case 'keys':
                    // map keys as a set
                    this.applyFun(app.id, 1, map => {
                        return (0, maybe_1.just)(runtimeValue_1.rv.mkSet(map.toMap().keys()));
                    });
                    break;
                case 'oneOf':
                    this.applyOneOf(app.id);
                    break;
                case 'exists':
                    this.mapLambdaThenReduce(app.id, set => runtimeValue_1.rv.mkBool(set.find(([result, _]) => result.toBool()) !== undefined));
                    break;
                case 'forall':
                    this.mapLambdaThenReduce(app.id, set => runtimeValue_1.rv.mkBool(set.find(([result, _]) => !result.toBool()) === undefined));
                    break;
                case 'map':
                    this.mapLambdaThenReduce(app.id, array => runtimeValue_1.rv.mkSet(array.map(([result, _]) => result)));
                    break;
                case 'filter':
                    this.mapLambdaThenReduce(app.id, arr => runtimeValue_1.rv.mkSet(arr.filter(([r, _]) => r.toBool()).map(([_, e]) => e)));
                    break;
                case 'select':
                    this.mapLambdaThenReduce(app.id, arr => runtimeValue_1.rv.mkList(arr.filter(([r, _]) => r.toBool()).map(([_, e]) => e)));
                    break;
                case 'mapBy':
                    this.mapLambdaThenReduce(app.id, arr => runtimeValue_1.rv.mkMap(arr.map(([v, k]) => [k, v])));
                    break;
                case 'Map':
                    this.applyFun(app.id, app.args.length, (...pairs) => (0, maybe_1.just)(runtimeValue_1.rv.mkMap(pairs)));
                    break;
                case 'setToMap':
                    this.applyFun(app.id, 1, (set) => (0, maybe_1.just)(runtimeValue_1.rv.mkMap(set.toSet().map(p => {
                        const arr = p.toList().toArray();
                        return [arr[0], arr[1]];
                    }))));
                    break;
                case 'setOfMaps':
                    this.applyFun(app.id, 2, (dom, rng) => {
                        return (0, maybe_1.just)(runtimeValue_1.rv.mkMapSet(dom, rng));
                    });
                    break;
                case 'then':
                    this.translateAllOrThen(app);
                    break;
                case 'fail':
                    this.applyFun(app.id, 1, result => {
                        return (0, maybe_1.just)(runtimeValue_1.rv.mkBool(!result.toBool()));
                    });
                    break;
                case 'expect':
                    this.translateExpect(app);
                    break;
                case 'assert':
                    this.applyFun(app.id, 1, cond => {
                        if (!cond.toBool()) {
                            this.errorTracker.addRuntimeError(app.id, 'QNT508', 'Assertion failed');
                            return (0, maybe_1.none)();
                        }
                        return (0, maybe_1.just)(cond);
                    });
                    break;
                case 'reps':
                    this.translateReps(app);
                    break;
                case 'q::test':
                    // the special operator that runs random simulation
                    this.test(app.id);
                    break;
                case 'q::testOnce':
                    // the special operator that runs random simulation
                    this.testOnce(app.id);
                    break;
                case 'q::debug':
                    this.applyFun(app.id, 2, (msg, value) => {
                        let columns = (0, graphics_1.terminalWidth)();
                        let valuePretty = (0, prettierimp_1.format)(columns, 0, (0, graphics_1.prettyQuintEx)(value.toQuintEx(idGenerator_1.zerog)));
                        console.log('>', msg.toStr(), valuePretty.toString());
                        return (0, maybe_1.just)(value);
                    });
                    break;
                // standard unary operators that are not handled by REPL
                case 'allLists':
                case 'chooseSome':
                case 'always':
                case 'eventually':
                case 'enabled':
                    this.applyFun(app.id, 1, _ => {
                        this.errorTracker.addRuntimeError(app.id, 'QNT501', `Runtime does not support the built-in operator '${app.opcode}'`);
                        return (0, maybe_1.none)();
                    });
                    break;
                // builtin operators that are not handled by REPL
                case 'orKeep':
                case 'mustChange':
                case 'weakFair':
                case 'strongFair':
                    this.applyFun(app.id, 2, _ => {
                        this.errorTracker.addRuntimeError(app.id, 'QNT501', `Runtime does not support the built-in operator '${app.opcode}'`);
                        return (0, maybe_1.none)();
                    });
                    break;
                default:
                    (0, util_1.unreachable)(app.opcode);
            }
        }
    }
    applyUserDefined(app) {
        const onError = (sourceId, msg) => {
            this.errorTracker.addCompileError(sourceId, 'QNT501', msg);
            this.compStack.push(runtime_1.fail);
        };
        // look up for the operator to see, whether it's just an operator, or a parameter
        const lookupEntry = this.lookupTable.get(app.id);
        if (lookupEntry === undefined) {
            return onError(app.id, `Called unknown operator ${app.opcode}`);
        }
        // this function gives us access to the compiled operator later
        let callableRef;
        if (lookupEntry === undefined || lookupEntry.kind !== 'param') {
            // The common case: the operator has been defined elsewhere.
            // We simply look up for the operator and return it via callableRef.
            const callable = this.contextLookup(app.id, ['callable']);
            if (callable === undefined || callable.nparams === undefined) {
                return onError(app.id, `Called unknown operator ${app.opcode}`);
            }
            callableRef = () => (0, maybe_1.just)(callable);
        }
        else {
            // The operator is a parameter of another operator.
            // We do not have access to the operator yet.
            let register = this.contextLookup(app.id, ['arg']);
            if (register === undefined) {
                return onError(app.id, `Parameter ${app.opcode} is not found`);
            }
            // every time we need a Callable, we retrieve it from the register
            callableRef = () => {
                return register.registerValue.map(v => v);
            };
        }
        const nargs = app.args.length; // === operScheme.type.args.length
        const nactual = this.compStack.length;
        if (nactual < nargs) {
            return onError(app.id, `Expected ${nargs} arguments for ${app.opcode}, found: ${nactual}`);
        }
        else {
            // pop nargs elements of the compStack
            const args = this.compStack.splice(-nargs, nargs);
            // Produce the new computable value.
            // This code is similar to applyFun, but it calls the listener before
            const comp = {
                eval: () => {
                    this.execListener.onUserOperatorCall(app);
                    // compute the values of the arguments at this point
                    const merged = (0, maybe_1.merge)(args.map(a => a.eval()));
                    const callable = callableRef();
                    if (merged.isNone() || callable.isNone()) {
                        this.execListener.onUserOperatorReturn(app, [], (0, maybe_1.none)());
                        return (0, maybe_1.none)();
                    }
                    return merged
                        .map(values => {
                        const result = callable.value.eval(values.map(maybe_1.just));
                        this.execListener.onUserOperatorReturn(app, values, result);
                        return result;
                    })
                        .join();
                },
            };
            this.compStack.push(comp);
        }
    }
    enterLambda(lam) {
        // introduce a register for every parameter
        lam.params.forEach(p => {
            const register = (0, runtime_1.mkRegister)('arg', p.name, (0, maybe_1.none)(), () => `Parameter ${p} is not set`);
            this.context.set((0, runtime_1.kindName)('arg', p.id), register);
            if (specialNames.includes(p.name)) {
                this.context.set((0, runtime_1.kindName)('arg', p.name), register);
            }
        });
        // After this point, the body of the lambda gets compiled.
        // The body of the lambda may refer to the parameter via names,
        // which are stored in the registers we've just created.
    }
    exitLambda(lam) {
        // The expression on the stack is the body of the lambda.
        // Transform it to Callable together with the registers.
        const registers = [];
        lam.params.forEach(p => {
            const id = specialNames.includes(p.name) ? p.name : p.id;
            const key = (0, runtime_1.kindName)('arg', id);
            const register = this.contextGet(id, ['arg']);
            if (register && register.registerValue) {
                this.context.delete(key);
                registers.push(register);
            }
            else {
                this.errorTracker.addCompileError(p.id, 'QNT501', `Parameter ${p.name} not found`);
            }
        });
        const lambdaBody = this.compStack.pop();
        if (lambdaBody) {
            this.compStack.push((0, runtime_1.mkCallable)(registers, lambdaBody));
        }
        else {
            this.errorTracker.addCompileError(lam.id, 'QNT501', 'Compilation of lambda failed');
        }
    }
    translateAssign(sourceId) {
        if (this.compStack.length < 2) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', `Assignment '=' needs two arguments`);
            return;
        }
        const [register, rhs] = this.compStack.splice(-2);
        const name = register.name;
        if (name === undefined) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', `Assignment '=' applied to a non-variable`);
            this.compStack.push(runtime_1.fail);
            return;
        }
        const nextvar = this.contextGet(name, ['nextvar']);
        if (nextvar) {
            this.compStack.push(rhs);
            this.applyFun(sourceId, 1, value => {
                nextvar.registerValue = (0, maybe_1.just)(value);
                return (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true));
            });
        }
        else {
            this.errorTracker.addCompileError(sourceId, 'QNT502', `Undefined next variable in ${name} = ...`);
            this.compStack.push(runtime_1.fail);
        }
    }
    /**
     * A generalized application of a one-argument Callable to a set-like
     * runtime value, as required by `exists`, `forall`, `map`, and `filter`.
     *
     * This method expects `compStack` to look like follows:
     *
     *  - `(top)` translated lambda, as `Callable`.
     *
     *  - `(top - 1)`: a set-like value to iterate over, as `Computable`.
     *
     * The method evaluates the Callable for each element of the iterable value
     * and either produces `none`, if evaluation failed for one of the elements,
     * or it applies `mapResultAndElems` to the pairs that consists of the Callable
     * result and the original element of the iterable value.
     * The final result is stored on the stack.
     */
    mapLambdaThenReduce(sourceId, mapResultAndElems) {
        if (this.compStack.length < 2) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', 'Not enough arguments');
            return;
        }
        // lambda translated to Callable
        const callable = this.compStack.pop();
        // apply the lambda to a single element of the set
        const evaluateElem = function (elem) {
            // evaluate the predicate against the actual arguments
            const result = callable.eval([(0, maybe_1.just)(elem)]);
            return result.map(result => [result, elem]);
        };
        this.applyFun(sourceId, 1, (iterable) => {
            return flatMap(iterable, evaluateElem).map(rs => mapResultAndElems(rs));
        });
    }
    /**
     * Translate one of the operators: fold, foldl, and foldr.
     */
    applyFold(sourceId, order) {
        if (this.compStack.length < 3) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', 'Not enough arguments for fold');
            return;
        }
        // extract two arguments from the call stack and keep the set
        const callable = this.compStack.pop();
        // this method supports only 2-argument callables
        (0, assert_1.strict)(callable.nparams === 2);
        // compile the computation of the initial value
        const initialComp = this.compStack.pop() ?? runtime_1.fail;
        // the register number depends on the traversal order
        const accumIndex = order == 'fwd' ? 0 : 1;
        // keep the iteration values in args
        const args = [(0, maybe_1.none)(), (0, maybe_1.none)()];
        // apply the lambda to a single element of the set
        const evaluateElem = function (elem) {
            // The accumulator should have been set in the previous iteration.
            // Set the other register to the element.
            args[1 - accumIndex] = (0, maybe_1.just)(elem);
            const result = callable.eval(args);
            // save the result for the next iteration
            args[accumIndex] = result;
            return result;
        };
        // iterate over the iterable (a set or a list)
        this.applyFun(sourceId, 1, (iterable) => {
            return initialComp
                .eval()
                .map(initialValue => {
                // save the initial value
                args[accumIndex] = (0, maybe_1.just)(initialValue);
                // fold the iterable
                return flatForEach(order, iterable, (0, maybe_1.just)(initialValue), evaluateElem);
            })
                .join();
        });
    }
    // pop nargs computable values, pass them the 'fun' function, and
    // push the combined computable value on the stack
    applyFun(sourceId, nargs, fun) {
        if (this.compStack.length < nargs) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', 'Not enough arguments');
        }
        else {
            // pop nargs elements of the compStack
            const args = this.compStack.splice(-nargs, nargs);
            // produce the new computable value
            const comp = {
                eval: () => {
                    try {
                        // compute the values of the arguments at this point
                        const values = args.map(a => a.eval());
                        // if they are all defined, apply the function 'fun' to the arguments
                        return (0, maybe_1.merge)(values)
                            .map(vs => fun(...vs.map(v => v)))
                            .join();
                    }
                    catch (error) {
                        const msg = error instanceof Error ? error.message : 'unknown error';
                        this.errorTracker.addRuntimeError(sourceId, 'QNT501', msg);
                        return (0, maybe_1.none)();
                    }
                },
            };
            this.compStack.push(comp);
        }
    }
    // if-then-else requires special treatment,
    // as it should not evaluate both arms
    translateIfThenElse(sourceId) {
        if (this.compStack.length < 3) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', 'Not enough arguments');
        }
        else {
            // pop 3 elements of the compStack
            const [cond, thenArm, elseArm] = this.compStack.splice(-3, 3);
            // produce the new computable value
            const comp = {
                eval: () => {
                    // compute the values of the arguments at this point
                    const v = cond.eval().map(pred => (pred.equals(runtimeValue_1.rv.mkBool(true)) ? thenArm.eval() : elseArm.eval()));
                    return v.join();
                },
            };
            this.compStack.push(comp);
        }
    }
    /**
     *  Compute all { ... } or A.then(B)...then(E) for a chain of actions.
     * @param actions actions as computable to execute
     * @param kind is it 'all { ... }' or 'A.then(B)'?
     * @param actionId given the action index, return the id that produced this action
     * @returns evaluation result
     */
    chainAllOrThen(actions, kind, actionId) {
        // save the values of the next variables, as actions may update them
        const savedValues = this.snapshotNextVars();
        const savedTrace = this.trace.get();
        let result = (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true));
        // Evaluate arguments iteratively.
        // Stop as soon as one of the arguments returns false.
        // This is a form of Boolean short-circuiting.
        let nactionsLeft = actions.length;
        for (const action of actions) {
            nactionsLeft--;
            result = action.eval();
            const isFalse = result.isJust() && !result.value.toBool();
            if (result.isNone() || isFalse) {
                // As soon as one of the arguments does not evaluate to true,
                // break out of the loop.
                // Restore the values of the next variables,
                // as evaluation was not successful.
                this.recoverNextVars(savedValues);
                this.trace.reset(savedTrace);
                if (kind === 'then' && nactionsLeft > 0 && isFalse) {
                    // Cannot extend a run. Emit an error message.
                    const actionNo = actions.length - (nactionsLeft + 1);
                    this.errorTracker.addRuntimeError(actionId(actionNo), 'QNT513', `Cannot continue in A.then(B), A evaluates to 'false'`);
                    return (0, maybe_1.none)();
                }
                else {
                    return result;
                }
            }
            // switch to the next frame, when implementing A.then(B)
            if (kind === 'then' && nactionsLeft > 0) {
                const oldState = this.varsToRecord();
                this.shiftVars();
                const newState = this.varsToRecord();
                this.trace.extend(newState);
                this.execListener.onNextState(oldState, newState);
            }
        }
        return result;
    }
    // translate all { A, ..., C } or A.then(B)
    translateAllOrThen(app) {
        if (this.compStack.length < app.args.length) {
            this.errorTracker.addCompileError(app.id, 'QNT501', `Not enough arguments on stack for "${app.opcode}"`);
            return;
        }
        const args = this.compStack.splice(-app.args.length);
        const kind = app.opcode === 'then' ? 'then' : 'all';
        const lazyCompute = () => this.chainAllOrThen(args, kind, idx => app.args[idx].id);
        this.compStack.push(mkFunComputable(lazyCompute));
    }
    // Translate A.expect(P):
    //  - Evaluate A.
    //  - When A's result is 'false', emit a runtime error.
    //  - When A's result is 'true':
    //    - Commit the variable updates: Shift the primed variables to unprimed.
    //    - Evaluate `P`.
    //    - If `P` evaluates to `false`, emit a runtime error (similar to `assert`).
    //    - If `P` evaluates to `true`, rollback to the previous state and return `true`.
    translateExpect(app) {
        // The code below is an adaption of chainAllOrThen.
        // If someone finds how to nicely combine both, please refactor.
        if (this.compStack.length !== 2) {
            this.errorTracker.addCompileError(app.id, 'QNT501', `Not enough arguments on stack for "${app.opcode}"`);
            return;
        }
        const [action, pred] = this.compStack.splice(-2);
        const lazyCompute = () => {
            const savedNextVars = this.snapshotNextVars();
            const savedTrace = this.trace.get();
            const actionResult = action.eval();
            if (actionResult.isNone() || !actionResult.value.toBool()) {
                // 'A' evaluates to 'false', or produces an error.
                // Restore the values of the next variables.
                this.recoverNextVars(savedNextVars);
                this.trace.reset(savedTrace);
                // expect emits an error when the run could not finish
                this.errorTracker.addRuntimeError(app.args[0].id, 'QNT508', 'Cannot continue to "expect"');
                return (0, maybe_1.none)();
            }
            else {
                const savedVarsAfterAction = this.snapshotVars();
                const savedNextVarsAfterAction = this.snapshotNextVars();
                const savedTraceAfterAction = this.trace.get();
                // Temporarily, switch to the next frame, to make a look-ahead evaluation.
                // For example, if `x == 1` and `x' == 2`, we would have `x == 2` and `x'` would be undefined.
                this.shiftVars();
                // evaluate P
                const predResult = pred.eval();
                // Recover the assignments to unprimed and primed variables.
                // E.g., we recover variables to `x == 1` and `x' == 2` in the above example.
                // This lets us combine the result of `expect` with other actions via `then`.
                // For example: `A.expect(P).then(B)`.
                this.recoverVars(savedVarsAfterAction);
                this.recoverNextVars(savedNextVarsAfterAction);
                this.trace.reset(savedTraceAfterAction);
                if (predResult.isNone() || !predResult.value.toBool()) {
                    this.errorTracker.addRuntimeError(app.args[1].id, 'QNT508', 'Expect condition does not hold true');
                    return (0, maybe_1.none)();
                }
                return predResult;
            }
        };
        this.compStack.push(mkFunComputable(lazyCompute));
    }
    // translate n.reps(A)
    translateReps(app) {
        if (this.compStack.length < 2) {
            this.errorTracker.addCompileError(app.id, 'QNT501', `Not enough arguments on stack for "${app.opcode}"`);
            return;
        }
        const [niterations, action] = this.compStack.splice(-2);
        const lazyCompute = () => {
            // compute the number of iterations and repeat 'action' that many times
            return niterations
                .eval()
                .map(num => {
                const n = Number(num.toInt());
                const indices = [...Array(n).keys()];
                const actions = indices.map(i => {
                    return {
                        eval: () => {
                            return action.eval([(0, maybe_1.just)(runtimeValue_1.rv.mkInt(BigInt(i)))]);
                        },
                    };
                });
                // In case the case of reps, we have multiple copies of the same action.
                // This is why all occurrences have the same id.
                return this.chainAllOrThen(actions, 'then', _ => app.args[1].id);
            })
                .join();
        };
        this.compStack.push(mkFunComputable(lazyCompute));
    }
    // translate one of the Boolean operators with short-circuiting:
    //  - or  { A, ..., C }
    //  - and { A, ..., C }
    //  - A implies B
    translateBoolOp(app, defaultValue, shortCircuit) {
        if (this.compStack.length < app.args.length) {
            this.errorTracker.addCompileError(app.id, 'QNT501', `Not enough arguments on stack for "${app.opcode}"`);
            return;
        }
        const args = this.compStack.splice(-app.args.length);
        const lazyCompute = () => {
            let result = defaultValue;
            // Evaluate arguments iteratively.
            // Stop as soon as shortCircuit tells us to stop.
            let no = 0;
            for (const arg of args) {
                // either the argument is evaluated to a Boolean, or fails
                result = arg.eval();
                if (result.isNone()) {
                    return (0, maybe_1.none)();
                }
                // if shortCircuit returns a value, return the value immediately
                const b = shortCircuit(no, result.value.toBool());
                if (b.isJust()) {
                    return b;
                }
                no += 1;
            }
            return result;
        };
        this.compStack.push(mkFunComputable(lazyCompute));
    }
    // translate any { A, ..., C }
    translateOrAction(app) {
        if (this.compStack.length < app.args.length) {
            this.errorTracker.addCompileError(app.id, 'QNT501', 'Not enough arguments on stack for "any"');
            return;
        }
        const args = this.compStack.splice(-app.args.length);
        // According to the semantics of action-level disjunctions,
        // we have to find out which branches are enabled and pick one of them
        // non-deterministically. Instead of modeling non-determinism,
        // we use a random number generator. This may change in the future.
        const lazyCompute = () => {
            // save the values of the next variables, as actions may update them
            const valuesBefore = this.snapshotNextVars();
            // we store the potential successor values in this array
            const successors = [];
            const successorIndices = [];
            // Evaluate arguments iteratively.
            args.forEach((arg, i) => {
                this.recoverNextVars(valuesBefore);
                // either the argument is evaluated to false, or fails
                this.execListener.onAnyOptionCall(app, i);
                const result = arg.eval().or((0, maybe_1.just)(runtimeValue_1.rv.mkBool(false)));
                const boolResult = result.unwrap().toBool();
                this.execListener.onAnyOptionReturn(app, i);
                // if this arm evaluates to true, save it in the candidates
                if (boolResult === true) {
                    successors.push(this.snapshotNextVars());
                    successorIndices.push(i);
                }
            });
            const ncandidates = successors.length;
            let choice;
            if (ncandidates === 0) {
                // no successor: restore the state and return false
                this.recoverNextVars(valuesBefore);
                this.execListener.onAnyReturn(args.length, -1);
                return (0, maybe_1.just)(runtimeValue_1.rv.mkBool(false));
            }
            else if (ncandidates === 1) {
                // There is exactly one successor, the execution is deterministic.
                // No need for randomization. This may reduce the number of tests.
                choice = 0;
            }
            else {
                // randomly pick a successor and return true
                choice = Number(this.rand(BigInt(ncandidates)));
            }
            this.recoverNextVars(successors[choice]);
            this.execListener.onAnyReturn(args.length, successorIndices[choice]);
            return (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true));
        };
        this.compStack.push(mkFunComputable(lazyCompute));
    }
    // Apply the operator oneOf
    applyOneOf(sourceId) {
        this.applyFun(sourceId, 1, set => {
            const bounds = set.bounds();
            const positions = bounds.map(b => {
                return (b
                    .map(sz => {
                    if (sz === 0n) {
                        this.errorTracker.addRuntimeError(sourceId, 'QNT509', `Applied oneOf to an empty set`);
                    }
                    return this.rand(sz);
                })
                    // An infinite set, pick an integer from the range [-2^255, 2^255).
                    // Note that pick on Nat uses the absolute value of the passed integer.
                    // TODO: make it a configurable parameter:
                    // https://github.com/informalsystems/quint/issues/279
                    .or((0, maybe_1.just)(-(2n ** 255n) + this.rand(2n ** 256n)))
                    .unwrap());
            });
            return set.pick(positions.values());
        });
    }
    test(sourceId) {
        if (this.compStack.length < 5) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', 'Not enough arguments on stack for "q::test"');
            return;
        }
        const [nruns, nsteps, init, next, inv] = this.compStack.splice(-5);
        this.runTestSimulation(nruns, nsteps, init, next, inv);
    }
    testOnce(sourceId) {
        if (this.compStack.length < 4) {
            this.errorTracker.addCompileError(sourceId, 'QNT501', 'Not enough arguments on stack for "q::testOnce"');
            return;
        }
        const [nsteps, init, next, inv] = this.compStack.splice(-4);
        const nruns = mkConstComputable(runtimeValue_1.rv.mkInt(1n));
        this.runTestSimulation(nruns, nsteps, init, next, inv);
    }
    // The simulator core: produce multiple random runs
    // and check the given state invariant (state assertion).
    //
    // Technically, this is similar to the implementation of folds.
    // However, it also restores the state and saves a trace, if there is any.
    runTestSimulation(nrunsComp, nstepsComp, init, next, inv) {
        const doRun = () => {
            return (0, maybe_1.merge)([nrunsComp, nstepsComp].map(c => c.eval()))
                .map(([nrunsRes, nstepsRes]) => {
                const isTrue = (res) => {
                    return !res.isNone() && res.value.toBool() === true;
                };
                // a failure flag for the case a runtime error is found
                let failure = false;
                // the value to be returned in the end of evaluation
                let errorFound = false;
                // save the registers to recover them later
                const vars = this.snapshotVars();
                const nextVars = this.snapshotNextVars();
                // do multiple runs, stop at the first failing run
                const nruns = nrunsRes.toInt();
                for (let runNo = 0; !errorFound && !failure && runNo < nruns; runNo++) {
                    this.execListener.onRunCall();
                    this.trace.reset();
                    // check Init()
                    const initApp = { id: 0n, kind: 'app', opcode: 'q::initAndInvariant', args: [] };
                    this.execListener.onUserOperatorCall(initApp);
                    const initResult = init.eval();
                    failure = initResult.isNone() || failure;
                    if (!isTrue(initResult)) {
                        this.execListener.onUserOperatorReturn(initApp, [], initResult);
                    }
                    else {
                        // The initial action evaluates to true.
                        // Our guess of values was good.
                        this.shiftVars();
                        this.trace.extend(this.varsToRecord());
                        // check the invariant Inv
                        const invResult = inv.eval();
                        this.execListener.onUserOperatorReturn(initApp, [], initResult);
                        failure = invResult.isNone() || failure;
                        if (!isTrue(invResult)) {
                            errorFound = true;
                        }
                        else {
                            // check all { Next(), shift(), Inv } in a loop
                            const nsteps = nstepsRes.toInt();
                            for (let i = 0; !errorFound && !failure && i < nsteps; i++) {
                                const nextApp = {
                                    id: 0n,
                                    kind: 'app',
                                    opcode: 'q::stepAndInvariant',
                                    args: [],
                                };
                                this.execListener.onUserOperatorCall(nextApp);
                                const nextResult = next.eval();
                                failure = nextResult.isNone() || failure;
                                if (isTrue(nextResult)) {
                                    this.shiftVars();
                                    this.trace.extend(this.varsToRecord());
                                    errorFound = !isTrue(inv.eval());
                                    this.execListener.onUserOperatorReturn(nextApp, [], nextResult);
                                }
                                else {
                                    // Otherwise, the run cannot be extended.
                                    // In some cases, this may indicate a deadlock.
                                    // Since we are doing random simulation, it is very likely
                                    // that we have not generated good values for extending
                                    // the run. Hence, do not report an error here, but simply
                                    // drop the run. Otherwise, we would have a lot of false
                                    // positives, which look like deadlocks but they are not.
                                    this.execListener.onUserOperatorReturn(nextApp, [], nextResult);
                                    this.execListener.onRunReturn((0, maybe_1.just)(runtimeValue_1.rv.mkBool(true)), this.trace.get());
                                    break;
                                }
                            }
                        }
                    }
                    const outcome = !failure ? (0, maybe_1.just)(runtimeValue_1.rv.mkBool(!errorFound)) : (0, maybe_1.none)();
                    this.execListener.onRunReturn(outcome, this.trace.get());
                    // recover the state variables
                    this.recoverVars(vars);
                    this.recoverNextVars(nextVars);
                } // end of a single random run
                // finally, return true, if no error was found
                return !failure ? (0, maybe_1.just)(runtimeValue_1.rv.mkBool(!errorFound)) : (0, maybe_1.none)();
            })
                .join();
        };
        this.compStack.push(mkFunComputable(doRun));
    }
    // convert the current variable values to a record
    varsToRecord() {
        const map = this.vars
            .filter(r => r.registerValue.isJust())
            .map(r => [r.name, r.registerValue.value]);
        return runtimeValue_1.rv.mkRecord(map);
    }
    shiftVars() {
        this.recoverVars(this.snapshotNextVars());
        this.nextVars.forEach(r => (r.registerValue = (0, maybe_1.none)()));
    }
    // save the values of the vars into an array
    snapshotVars() {
        return this.vars.map(r => r.registerValue);
    }
    // save the values of the next vars into an array
    snapshotNextVars() {
        return this.nextVars.map(r => r.registerValue);
    }
    // load the values of the variables from an array
    recoverVars(values) {
        this.vars.forEach((r, i) => (r.registerValue = values[i]));
    }
    // load the values of the next variables from an array
    recoverNextVars(values) {
        this.nextVars.forEach((r, i) => (r.registerValue = values[i]));
    }
    contextGet(name, kinds) {
        for (const k of kinds) {
            const value = this.context.get((0, runtime_1.kindName)(k, name));
            if (value) {
                return value;
            }
        }
        return undefined;
    }
    contextLookup(id, kinds) {
        const vdef = this.lookupTable.get(id);
        if (vdef) {
            const refId = vdef.id;
            for (const k of kinds) {
                const value = this.context.get((0, runtime_1.kindName)(k, refId));
                if (value) {
                    return value;
                }
            }
        }
        return undefined;
    }
    // Access a list via an index
    getListElem(sourceId, list, idx) {
        if (idx >= 0n && idx < list.size) {
            const elem = list.get(Number(idx));
            return elem ? (0, maybe_1.just)(elem) : (0, maybe_1.none)();
        }
        else {
            this.errorTracker.addRuntimeError(sourceId, 'QNT510', `Out of bounds, nth(${idx})`);
            return (0, maybe_1.none)();
        }
    }
    // Update a list via an index
    updateList(sourceId, list, idx, value) {
        if (idx >= 0n && idx < list.size) {
            return (0, maybe_1.just)(runtimeValue_1.rv.mkList(list.set(Number(idx), value)));
        }
        else {
            this.errorTracker.addRuntimeError(sourceId, 'QNT510', `Out of bounds, replaceAt(..., ${idx}, ...)`);
            return (0, maybe_1.none)();
        }
    }
    // slice a list
    sliceList(sourceId, list, start, end) {
        return (0, maybe_1.just)(runtimeValue_1.rv.mkList(list.slice(start, end)));
    }
}
exports.CompilerVisitor = CompilerVisitor;
// make a `Computable` that always returns a given runtime value
function mkConstComputable(value) {
    return {
        eval: () => {
            return (0, maybe_1.just)(value);
        },
    };
}
// make a `Computable` that always returns a given runtime value
function mkFunComputable(fun) {
    return {
        eval: () => {
            return fun();
        },
    };
}
/**
 * Apply `f` to every element of `iterable` and either:
 *
 *  - return `none`, if one of the results is `none`, or
 *  - return `just` of the unpacked results.
 */
function flatMap(iterable, f) {
    const results = [];
    for (const arg of iterable) {
        const res = f(arg);
        if (res.isNone()) {
            return (0, maybe_1.none)();
        }
        else {
            const { value } = res;
            results.push(value);
        }
    }
    return (0, maybe_1.just)(results);
}
/**
 * Apply `f` to every element of `iterable` and either:
 *
 *  - return `none`, if one of the results is `none`, or
 *  - return `just` of the last result.
 */
function flatForEach(order, iterable, init, f) {
    let result = init;
    // if the reverse order is required, reverse the array
    const iter = order === 'fwd' ? iterable : Array(...iterable).reverse();
    for (const arg of iter) {
        result = f(arg);
        if (result.isNone()) {
            return result;
        }
    }
    return result;
}
//# sourceMappingURL=compilerImpl.js.map