import { Maybe } from '@sweet-monads/maybe';
import { ValueObject } from 'immutable';
import { QuintEx } from '../ir/quintIr';
import { IdGenerator } from '../idGenerator';
/**
 * Evaluation result.
 * The implementation details are hidden behind this interface.
 */
export interface EvalResult extends ValueObject {
    /**
     * Convert an evaluation result to Quint.
     *
     * This function always returns sets in the normalized representation,
     * that is, in `set(elements)`, the elements are ordered according to their
     * string representation. As sorting via strings may be slow, we do not
     * recommend using `toQuintEx` in computation-intensive code.
     *
     * @param gen a generator that produces unique ids
     * @return this evaluation result converted to Quint expression.
     */
    toQuintEx(gen: IdGenerator): QuintEx;
}
/**
 * An object that can be evaluated by the runtime. Normally, it is constructed
 * from a Quint expression, but it does not have to.
 */
export interface Computable {
    /**
     * The simplest form of evaluation. Just compute the value.
     * This method may return none, if a computation error happens during
     * evaluation.
     *
     * @param args optional arguments to the computable
     */
    eval: (args?: Maybe<any>[]) => Maybe<EvalResult>;
}
/**
 * The kind of a computable.
 */
export type ComputableKind = 'var' | 'nextvar' | 'arg' | 'callable';
/**
 * Create a key that encodes its name and kind. This is only useful for
 * storing registers in a map, as JS objects are hard to use as keys.
 * In a good language, we would use (kind, name), but not in JS.
 */
export declare const kindName: (kind: ComputableKind, name: string | bigint) => string;
/**
 * A computable that evaluates to the value of a readable/writeable register.
 */
export interface Register extends Computable {
    name: string;
    kind: ComputableKind;
    registerValue: Maybe<any>;
}
/**
 * Create an object that implements Register.
 */
export declare function mkRegister(kind: ComputableKind, registerName: string, initValue: Maybe<any>, onUndefined: () => void): Register;
/**
 * A callable value like an operator definition. Its body is computable.
 * One has to pass the values of the parameters to eval.
 */
export interface Callable extends Computable {
    /**
     * The number of parameters expected by the callable value.
     */
    nparams: number;
}
/**
 * Create an object that implements Callable.
 */
export declare function mkCallable(registers: Register[], body: Computable): Callable;
/**
 * An implementation of Computable that always fails.
 */
export declare const fail: {
    eval: () => Maybe<EvalResult>;
};
/**
 * An error produced during execution.
 */
export interface ExecError {
    msg: string;
    sourceAndLoc: string | undefined;
}
/**
 * An error handler that is called on any kind of error that is happening
 * during execution.
 */
export type ExecErrorHandler = (_error: ExecError) => void;
