import { Either } from '@sweet-monads/either';
import { QuintEx } from '../ir/quintIr';
import { CompilationState } from './compile';
import { LookupTable } from '../names/base';
import { ExecutionFrame } from './trace';
import { Rng } from '../rng';
import { QuintError } from '../quintError';
/**
 * Various settings to be passed to the testing framework.
 */
export interface TestOptions {
    testMatch: (n: string) => boolean;
    maxSamples: number;
    rng: Rng;
    verbosity: number;
    onTrace(index: number, name: string, status: string, vars: string[], states: QuintEx[]): void;
}
/**
 * Evaluation result.
 * The implementation details are hidden behind this interface.
 */
export interface TestResult {
    /**
     * The test name.
     */
    name: string;
    /**
     * The test status.
     */
    status: 'passed' | 'failed' | 'ignored';
    /**
     * The seed value to repeat the test.
     */
    seed: bigint;
    /**
     * When status === 'failed', errors contain the explanatory errors.
     */
    errors: QuintError[];
    /**
     * If the trace was recorded, frames contains the history.
     */
    frames: ExecutionFrame[];
    /**
     * The number of tried samples.
     */
    nsamples: number;
}
/**
 * Run a test suite of a single module.
 *
 * @param modules Quint modules in the intermediate representation
 * @param main the module that should be used as a state machine
 * @param sourceMap source map as produced by the parser
 * @param lookupTable lookup table as produced by the parser
 * @param analysisOutput the maps produced by the static analysis
 * @param options misc test options
 * @returns the `right(results)` of running the tests if the tests could be
 *    run, or `left(errors)` with any compilation or analysis errors that
 *    prevent the tests from running
 */
export declare function compileAndTest(compilationState: CompilationState, mainName: string, lookupTable: LookupTable, options: TestOptions): Either<QuintError[], TestResult[]>;
