"use strict";
/*
 * Unit-testing framework for the JS runtime.
 *
 * Igor Konnov, 2023
 *
 * Copyright 2022 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileAndTest = void 0;
const either_1 = require("@sweet-monads/either");
const compile_1 = require("./compile");
const idGenerator_1 = require("./../idGenerator");
const runtime_1 = require("./runtime");
const trace_1 = require("./trace");
const compilerImpl_1 = require("./impl/compilerImpl");
/**
 * Run a test suite of a single module.
 *
 * @param modules Quint modules in the intermediate representation
 * @param main the module that should be used as a state machine
 * @param sourceMap source map as produced by the parser
 * @param lookupTable lookup table as produced by the parser
 * @param analysisOutput the maps produced by the static analysis
 * @param options misc test options
 * @returns the `right(results)` of running the tests if the tests could be
 *    run, or `left(errors)` with any compilation or analysis errors that
 *    prevent the tests from running
 */
function compileAndTest(compilationState, mainName, lookupTable, options) {
    const main = compilationState.modules.find(m => m.name === mainName);
    if (!main) {
        return (0, either_1.left)([{ code: 'QNT405', message: `Main module ${mainName} not found` }]);
    }
    const recorder = (0, trace_1.newTraceRecorder)(options.verbosity, options.rng);
    const ctx = (0, compile_1.compile)(compilationState, (0, compilerImpl_1.newEvaluationState)(recorder), lookupTable, options.rng.next, main.declarations);
    const ctxErrors = ctx.syntaxErrors.concat(ctx.compileErrors, ctx.analysisErrors);
    if (ctxErrors.length > 0) {
        // In principle, these errors should have been caught earlier.
        // But if they did not, return immediately.
        return (0, either_1.left)(ctxErrors);
    }
    const saveTrace = (index, name, status) => {
        // Save the best traces that are reported by the recorder:
        // If a test failed, it is the first failing trace.
        // Otherwise, it is the longest trace explored by the test.
        const states = recorder.getBestTrace().args.map(e => e.toQuintEx(idGenerator_1.zerog));
        options.onTrace(index, name, status, ctx.evaluationState.vars.map(v => v.name), states);
    };
    const testDefs = main.declarations.filter(d => d.kind === 'def' && options.testMatch(d.name));
    return (0, either_1.mergeInMany)(testDefs.map((def, index) => {
        return getComputableForDef(ctx, def).map(comp => {
            recorder.clear();
            const name = def.name;
            // save the initial seed
            let seed = options.rng.getState();
            let nsamples = 1;
            // run up to maxSamples, stop on the first failure
            for (; nsamples <= options.maxSamples; nsamples++) {
                // record the seed value
                seed = options.rng.getState();
                recorder.onRunCall();
                // reset the trace
                ctx.evaluationState.trace.reset();
                // run the test
                const result = comp.eval();
                // extract the trace
                const trace = ctx.evaluationState.trace.get();
                if (trace.length > 0) {
                    recorder.onRunReturn(result, trace);
                }
                else {
                    // Report a non-critical error
                    console.error('Missing a trace');
                    recorder.onRunReturn(result, []);
                }
                // evaluate the result
                if (result.isNone()) {
                    // if the test failed, return immediately
                    return {
                        name,
                        status: 'failed',
                        errors: ctx.getRuntimeErrors(),
                        seed,
                        frames: recorder.getBestTrace().subframes,
                        nsamples: nsamples,
                    };
                }
                const ex = result.value.toQuintEx(idGenerator_1.zerog);
                if (ex.kind !== 'bool') {
                    // if the test returned a malformed result, return immediately
                    return {
                        name,
                        status: 'ignored',
                        errors: [],
                        seed: seed,
                        frames: recorder.getBestTrace().subframes,
                        nsamples: nsamples,
                    };
                }
                if (!ex.value) {
                    // if the test returned false, return immediately
                    const error = {
                        code: 'QNT511',
                        message: `Test ${name} returned false`,
                        reference: def.id,
                    };
                    saveTrace(index, name, 'failed');
                    return {
                        name,
                        status: 'failed',
                        errors: [error],
                        seed: seed,
                        frames: recorder.getBestTrace().subframes,
                        nsamples: nsamples,
                    };
                }
                else {
                    if (options.rng.getState() === seed) {
                        // This successful test did not use non-determinism.
                        // Running it one time is sufficient.
                        saveTrace(index, name, 'passed');
                        return {
                            name,
                            status: 'passed',
                            errors: [],
                            seed: seed,
                            frames: recorder.getBestTrace().subframes,
                            nsamples: nsamples,
                        };
                    }
                }
            }
            // the test was run maxSamples times, and no errors were found
            saveTrace(index, name, 'passed');
            return {
                name,
                status: 'passed',
                errors: [],
                seed: seed,
                frames: recorder.getBestTrace().subframes,
                nsamples: nsamples - 1,
            };
        });
    }));
}
exports.compileAndTest = compileAndTest;
function getComputableForDef(ctx, def) {
    const comp = ctx.evaluationState.context.get((0, runtime_1.kindName)('callable', def.id));
    if (comp) {
        return (0, either_1.right)(comp);
    }
    else {
        return (0, either_1.left)({ code: 'QNT501', message: `Cannot find computable for ${def.name}`, reference: def.id });
    }
}
//# sourceMappingURL=testing.js.map