"use strict";
/*
 * Data structures for recording execution traces.
 *
 * Igor Konnov, 2023
 *
 * Copyright 2022-2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.newTraceRecorder = exports.noExecutionListener = void 0;
const maybe_1 = require("@sweet-monads/maybe");
const assert_1 = require("assert");
const verbosity_1 = require("./../verbosity");
const runtimeValue_1 = require("./impl/runtimeValue");
/**
 * An implementation of ExecutionListener that does nothing.
 */
exports.noExecutionListener = {
    onUserOperatorCall: (_app) => { },
    onUserOperatorReturn: (_app, _args, _result) => { },
    onAnyOptionCall: (_anyExpr, _position) => { },
    onAnyOptionReturn: (_anyExpr, _position) => { },
    onAnyReturn: (_noptions, _choice) => { },
    onNextState: (_oldState, _newState) => { },
    onRunCall: () => { },
    onRunReturn: (_outcome, _trace) => { },
};
// a trace recording listener
const newTraceRecorder = (verbosityLevel, rng) => {
    return new TraceRecorderImpl(verbosityLevel, rng);
};
exports.newTraceRecorder = newTraceRecorder;
// a private implementation of a trace recorder
class TraceRecorderImpl {
    constructor(verbosityLevel, rng) {
        this.verbosityLevel = verbosityLevel;
        this.rng = rng;
        const bottom = this.newBottomFrame();
        this.bestTrace = bottom;
        this.bestTraceSeed = rng.getState();
        this.runSeed = this.bestTraceSeed;
        this.frameStack = [bottom];
    }
    clear() {
        const bottom = this.newBottomFrame();
        this.bestTrace = bottom;
        this.frameStack = [bottom];
    }
    getBestTrace() {
        return this.bestTrace;
    }
    getBestTraceSeed() {
        return this.bestTraceSeed;
    }
    onUserOperatorCall(app) {
        // For now, we cannot tell apart actions from other user definitions.
        // https://github.com/informalsystems/quint/issues/747
        if (verbosity_1.verbosity.hasUserOpTracking(this.verbosityLevel)) {
            const newFrame = { app: app, args: [], result: (0, maybe_1.none)(), subframes: [] };
            if (this.frameStack.length == 0) {
                // this should not happen, as there is always bottomFrame,
                // but we do not throw here, as trace collection is not the primary
                // function of the simulator.
                this.frameStack = [newFrame];
            }
            else if (this.frameStack.length === 2 && this.frameStack[1].app.opcode === '_') {
                // A placeholder frame created from `q::input` or `then`. Modify in place.
                const frame = this.frameStack[1];
                frame.app = app;
                frame.args = [];
                frame.result = (0, maybe_1.none)();
                frame.subframes = [];
            }
            else {
                // connect the new frame to the previous frame
                this.frameStack[this.frameStack.length - 1].subframes.push(newFrame);
                // and push the new frame to be on top of the stack
                this.frameStack.push(newFrame);
            }
        }
    }
    onUserOperatorReturn(_app, args, result) {
        if (verbosity_1.verbosity.hasUserOpTracking(this.verbosityLevel)) {
            const top = this.frameStack.pop();
            if (top) {
                // since this frame is connected via the parent frame,
                // the result will not disappear
                top.args = args;
                top.result = result;
            }
        }
    }
    onAnyOptionCall(anyExpr, _position) {
        if (verbosity_1.verbosity.hasActionTracking(this.verbosityLevel)) {
            // the option has to be hidden under its own frame,
            // so it can be popped as a single frame later
            const newFrame = {
                app: anyExpr,
                args: [],
                result: (0, maybe_1.none)(),
                subframes: [],
            };
            if (this.frameStack.length > 0) {
                // add the option directly to the stack of the current expression
                // that contains `any { ... }`.
                this.frameStack[this.frameStack.length - 1].subframes.push(newFrame);
                this.frameStack.push(newFrame);
            }
            else {
                this.frameStack = [newFrame];
            }
        }
    }
    onAnyOptionReturn(_anyExpr, _position) {
        if (verbosity_1.verbosity.hasActionTracking(this.verbosityLevel)) {
            this.frameStack.pop();
        }
    }
    onAnyReturn(noptions, choice) {
        if (verbosity_1.verbosity.hasActionTracking(this.verbosityLevel)) {
            (0, assert_1.strict)(this.frameStack.length > 0);
            const top = this.frameStack[this.frameStack.length - 1];
            const start = top.subframes.length - noptions;
            // leave only the chosen frame as well as the older ones
            top.subframes = top.subframes.filter((_, i) => i < start || i == start + choice);
            if (choice >= 0) {
                // The top frame contains the frames of the chosen option that are
                // wrapped in anyExpr.args[position], see onAnyOptionCall.
                // Unwrap the option, as we do not need it any longer.
                const optionFrame = top.subframes.pop();
                if (optionFrame) {
                    top.subframes = top.subframes.concat(optionFrame.subframes);
                }
            }
        }
    }
    onNextState(_oldState, _newState) {
        // introduce a new frame that is labelled with a dummy operator
        if (verbosity_1.verbosity.hasUserOpTracking(this.verbosityLevel)) {
            const dummy = { id: 0n, kind: 'app', opcode: '_', args: [] };
            const newFrame = { app: dummy, args: [], result: (0, maybe_1.none)(), subframes: [] };
            // forget the frames, except the bottom one, and push the new one
            this.frameStack = [this.frameStack[0], newFrame];
            // connect the new frame to the topmost frame, which effects in a new step
            this.frameStack[0].subframes.push(newFrame);
        }
    }
    onRunCall() {
        // reset the stack
        this.frameStack = [this.newBottomFrame()];
        this.runSeed = this.rng.getState();
    }
    onRunReturn(outcome, trace) {
        (0, assert_1.strict)(this.frameStack.length > 0);
        const bottom = this.frameStack[0];
        const fromResult = (r) => {
            if (r.isNone()) {
                return true;
            }
            else {
                const rex = r.value.toQuintEx({ nextId: () => 0n });
                return rex.kind === 'bool' && !rex.value;
            }
        };
        const notOk = fromResult(outcome);
        const prevNotOk = fromResult(this.bestTrace.result);
        // Prefer short traces for error, and longer traces for non error.
        // Therefore, override the best trace only if:
        //  - there is an error, the new trace is shorter, or the old trace is non-error;
        //  - there is no error, the new trace is longer, and there was no error before.
        const override = notOk
            ? this.bestTrace.args.length === 0 || !prevNotOk || this.bestTrace.args.length >= bottom.args.length
            : !prevNotOk && this.bestTrace.args.length <= bottom.args.length;
        if (override) {
            this.bestTrace = bottom;
            this.bestTraceSeed = this.runSeed;
            this.bestTrace.result = outcome;
            this.bestTrace.args = trace;
        }
    }
    // create a bottom frame, which encodes the whole trace
    newBottomFrame() {
        return {
            // this is just a dummy operator application
            app: { id: 0n, kind: 'app', opcode: 'Rec', args: [] },
            // we will store the sequence of states here
            args: [],
            // the result of the trace evaluation
            result: (0, maybe_1.just)(runtimeValue_1.rv.mkBool(true)),
            // and here we store the subframes for the top-level actions
            subframes: [],
        };
    }
}
//# sourceMappingURL=trace.js.map