"use strict";
/*
 * A state-machine simulator.
 *
 * Igor Konnov, 2023
 *
 * Copyright 2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileAndRun = void 0;
const compile_1 = require("./runtime/compile");
const trace_1 = require("./runtime/trace");
const cliCommands_1 = require("./cliCommands");
const errorReporter_1 = require("./errorReporter");
const assert_1 = __importDefault(require("assert"));
function errSimulationResult(errors) {
    return {
        outcome: { status: 'error', errors },
        vars: [],
        states: [],
        frames: [],
        seed: 0n,
    };
}
/**
 * Execute a run.
 *
 * @param idGen a unique generator of identifiers
 * @param code the source code of the modules
 * @param mainStart the start index of the main module in the code
 * @param mainEnd the end index of the main module in the code
 * @param mainName the module that should be used as a state machine
 * @param mainPath the lookup path that was used to retrieve the main module
 * @param options simulator settings
 * @returns either error messages (left),
    or the trace as an expression (right)
 */
function compileAndRun(idGen, code, mainStart, mainEnd, mainName, mainPath, options) {
    // Once we have 'import from ...' implemented, we should pass
    // a filename instead of the source code (see #8)
    // Parse the code once again, but this time include the special definitions
    // that are required by the runner.
    // This code should be revisited in #618.
    const o = options;
    // Defs required by the simulator, to be added to the main module before compilation
    const extraDefs = [
        `def q::test(q::nrunsArg, q::nstepsArg, q::initArg, q::nextArg, q::invArg) = false`,
        `action q::init = { ${o.init} }`,
        `action q::step = { ${o.step} }`,
        `val q::inv = { ${o.invariant} }`,
        `val q::runResult = q::test(${o.maxSamples}, ${o.maxSteps}, q::init, q::step, q::inv)`,
    ];
    // Construct the modules' code, adding the extra definitions to the main module
    const newMainModuleCode = code.slice(mainStart, mainEnd - 1) + '\n' + extraDefs.join('\n');
    const codeWithExtraDefs = code.slice(0, mainStart) + newMainModuleCode + code.slice(mainEnd);
    const recorder = (0, trace_1.newTraceRecorder)(options.verbosity, options.rng);
    const ctx = (0, compile_1.compileFromCode)(idGen, codeWithExtraDefs, mainName, mainPath, recorder, options.rng.next);
    const compilationErrors = ctx.syntaxErrors.concat(ctx.analysisErrors).concat(ctx.compileErrors);
    if (compilationErrors.length > 0) {
        return errSimulationResult(compilationErrors);
    }
    // evaluate q::runResult, which triggers the simulator
    const evaluationState = ctx.evaluationState;
    const res = (0, compile_1.contextNameLookup)(evaluationState.context, 'q::runResult', 'callable');
    if (res.isLeft()) {
        const errors = [{ code: 'QNT512', message: res.value }];
        return errSimulationResult(errors);
    }
    else {
        res.value.eval();
    }
    const topFrame = recorder.getBestTrace();
    // Validate required outcome of correct simulation
    const maybeEvalResult = topFrame.result;
    (0, assert_1.default)(maybeEvalResult.isJust(), 'invalid simulation failed to produce a result');
    const quintExResult = maybeEvalResult.value.toQuintEx(idGen);
    (0, assert_1.default)(quintExResult.kind === 'bool', 'invalid simulation produced non-boolean value ');
    const simulationSucceeded = quintExResult.value;
    const vars = evaluationState.vars.map(v => v.name);
    const states = topFrame.args.map(e => e.toQuintEx(idGen));
    const frames = topFrame.subframes;
    const seed = recorder.getBestTraceSeed();
    const runtimeErrors = ctx.getRuntimeErrors();
    if (runtimeErrors.length > 0) {
        // FIXME(#1052) we shouldn't need to do this if the error id was not some non-sense generated in `compileFromCode`
        // The evaluated code source is not included in the context, so we crete a version with it for the error reporter
        const code = new Map([...ctx.compilationState.sourceCode.entries(), [mainPath.normalizedPath, codeWithExtraDefs]]);
        const finders = (0, errorReporter_1.createFinders)(code);
        const locatedErrors = runtimeErrors.map(error => ({
            code: error.code,
            // Include the location information (locs) in the error message - this
            // is the hacky part, as it should only be included at the CLI level
            message: (0, errorReporter_1.formatError)(code, finders, {
                // This will create the `locs` attribute and an explanation
                ...(0, cliCommands_1.mkErrorMessage)(ctx.compilationState.sourceMap)(error),
                // We override the explanation to keep the original one to avoid
                // duplication, since `mkErrorMessage` will be called again at the CLI
                // level. `locs` won't be touched then because this error doesn't
                // include a `reference` field
                explanation: error.message,
            }),
        }));
        // This should be kept after the hack is removed
        return {
            vars,
            states,
            frames,
            seed,
            outcome: { status: 'error', errors: locatedErrors },
        };
    }
    const status = simulationSucceeded ? 'ok' : 'violation';
    return {
        vars,
        states,
        frames,
        seed,
        outcome: { status },
    };
}
exports.compileAndRun = compileAndRun;
//# sourceMappingURL=simulation.js.map