"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mocha_1 = require("mocha");
const chai_1 = require("chai");
const either_1 = require("@sweet-monads/either");
const path_1 = require("path");
const sourceResolver_1 = require("../../src/parsing/sourceResolver");
const fs_1 = require("fs");
const eol_1 = require("eol");
const basename = (0, path_1.resolve)(__dirname, '../../testFixture');
// read a Quint file from the test data directory
function readQuint(name) {
    const p = (0, path_1.resolve)(__dirname, '../../testFixture', name + '.qnt');
    const content = (0, fs_1.readFileSync)(p).toString('utf8');
    return (0, eol_1.lf)(content);
}
(0, mocha_1.describe)('resolve sources from files', () => {
    (0, mocha_1.it)('parses empty module', () => {
        const expected = readQuint('./_0001emptyModule');
        const r = (0, sourceResolver_1.fileSourceResolver)();
        const result = r.load(r.lookupPath(basename, './_0001emptyModule.qnt'));
        chai_1.assert.deepEqual(result, (0, either_1.right)(expected));
    });
    (0, mocha_1.it)('parses via ..', () => {
        const expected = readQuint('./_0001emptyModule');
        const r = (0, sourceResolver_1.fileSourceResolver)();
        const result = r.load(r.lookupPath(basename, '../testFixture/_0001emptyModule.qnt'));
        chai_1.assert.deepEqual(result, (0, either_1.right)(expected));
    });
    (0, mocha_1.it)('errors on non-existant', () => {
        const r = (0, sourceResolver_1.fileSourceResolver)();
        const result = r.load(r.lookupPath(basename, 'does-not-exist'));
        chai_1.assert.deepEqual(result, (0, either_1.left)(`ENOENT: no such file or directory, open '${basename}/does-not-exist'`));
    });
    (0, mocha_1.it)('stemname', () => {
        const r = (0, sourceResolver_1.fileSourceResolver)();
        const result = r.stempath(r.lookupPath(`${basename}/testFixture`, './_0001emptyModule.qnt'));
        chai_1.assert.deepEqual(result, `${basename}/testFixture`);
    });
});
// a static table that contains sources in strings
const staticTable = new Map([
    ['/foo.qnt', 'module foo {}'],
    ['/bar.qnt', 'module bar {}'],
    ['/baz/baz.qnt', 'module baz {}'],
]);
(0, mocha_1.describe)('resolve sources from strings', () => {
    (0, mocha_1.it)('parses foo', () => {
        const expected = staticTable.get('/foo.qnt');
        const r = (0, sourceResolver_1.stringSourceResolver)(staticTable);
        const result = r.load(r.lookupPath('/', './foo.qnt'));
        chai_1.assert.deepEqual(result, (0, either_1.right)(expected));
    });
    (0, mocha_1.it)('parses bar via ..', () => {
        const expected = staticTable.get('/bar.qnt');
        const r = (0, sourceResolver_1.stringSourceResolver)(staticTable);
        const result = r.load(r.lookupPath('/', './baz/../bar.qnt'));
        chai_1.assert.deepEqual(result, (0, either_1.right)(expected));
    });
    (0, mocha_1.it)('parses baz via baz/baz.qnt', () => {
        const expected = staticTable.get('/baz/baz.qnt');
        const r = (0, sourceResolver_1.stringSourceResolver)(staticTable);
        const result = r.load(r.lookupPath('/', './baz/baz.qnt'));
        chai_1.assert.deepEqual(result, (0, either_1.right)(expected));
    });
    (0, mocha_1.it)('parses baz via baz.qnt', () => {
        const expected = staticTable.get('/baz/baz.qnt');
        const r = (0, sourceResolver_1.stringSourceResolver)(staticTable);
        const result = r.load(r.lookupPath('/baz', './baz.qnt'));
        chai_1.assert.deepEqual(result, (0, either_1.right)(expected));
    });
    (0, mocha_1.it)('errors on non-existant', () => {
        const r = (0, sourceResolver_1.stringSourceResolver)(staticTable);
        const result = r.load(r.lookupPath('/', 'does-not-exist'));
        chai_1.assert.deepEqual(result, (0, either_1.left)(`Source not found: '/does-not-exist'`));
    });
    (0, mocha_1.it)('stemname', () => {
        const r = (0, sourceResolver_1.stringSourceResolver)(staticTable);
        const result = r.stempath(r.lookupPath('/baz', './baz.qnt'));
        chai_1.assert.deepEqual(result, '/baz');
    });
});
//# sourceMappingURL=sourceResolver.test.js.map